# This file is part of PywerView.

# PywerView is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PywerView is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PywerView.  If not, see <http://www.gnu.org/licenses/>.

# Yannick Méheut [yannick (at) meheut (dot) org] - Copyright © 2023

from bs4 import BeautifulSoup
from io import BytesIO

from impacket.smbconnection import SMBConnection, SessionError

from pywerview.objects.adobjects import *
from pywerview.requester import LDAPRequester
from pywerview.functions.net import NetRequester

class GPORequester(LDAPRequester):

    @LDAPRequester._ldap_connection_init
    def get_netgpo(self, queried_gponame='*', queried_displayname=str(),
                   queried_domain=str(), ads_path=str()):

        gpo_search_filter = '(objectCategory=groupPolicyContainer)'

        if queried_displayname:
            gpo_search_filter += '(displayname={})'.format(queried_displayname)
        else:
            gpo_search_filter += '(name={})'.format(queried_gponame)

        gpo_search_filter = '(&{})'.format(gpo_search_filter)

        return self._ldap_search(gpo_search_filter, GPO)

    @LDAPRequester._ldap_connection_init
    def get_netpso(self, queried_psoname='*', queried_displayname=str(),
                   queried_domain=str(), ads_path=str()):

        pso_search_filter = '(objectClass=msDS-PasswordSettings)'

        if queried_displayname:
            pso_search_filter += '(displayname={})'.format(queried_displayname)
        else:
            pso_search_filter += '(name={})'.format(queried_psoname)

        pso_search_filter = '(&{})'.format(pso_search_filter)

        return self._ldap_search(pso_search_filter, PSO)

    def get_gpttmpl(self, gpttmpl_path):
        content_io = BytesIO()

        gpttmpl_path_split = gpttmpl_path.split('\\')
        target = self._domain_controller
        share = gpttmpl_path_split[3]
        file_name = '\\'.join(gpttmpl_path_split[4:])

        smb_connection = SMBConnection(remoteName=target, remoteHost=target)
        if self._do_kerberos:
            smb_connection.kerberosLogin(self._user, self._password, self._domain,
                                 self._lmhash, self._nthash)
        else:
            smb_connection.login(self._user, self._password, self._domain,
                                 self._lmhash, self._nthash)

        self._logger.debug('Get File: Share = {0}, file_name ={1}'.format(share, file_name))
        smb_connection.connectTree(share)
        smb_connection.getFile(share, file_name, content_io.write)
        try:
            content = content_io.getvalue().decode('utf-16le')[1:].replace('\r', '')
        except UnicodeDecodeError:
            self._logger.warning('Unicode error: trying utf-8')
            content = content_io.getvalue().decode('utf-8').replace('\r', '')

        gpttmpl_final = GptTmpl(list())
        for l in content.split('\n'):
            if l.startswith('['):
                section_name = l.strip('[]').replace(' ', '').lower()
                gpttmpl_final._attributes_dict[section_name] = Policy(list())
            elif '=' in l:
                property_name, property_values = [x.strip() for x in l.split('=')]
                if ',' in property_values:
                    property_values = property_values.split(',')
                gpttmpl_final._attributes_dict[section_name]._attributes_dict[property_name] = property_values

        return gpttmpl_final

    def get_domainpolicy(self, source='domain', queried_domain=str(),
                         resolve_sids=False):
        if source == 'domain':
            queried_gponame = '{31B2F340-016D-11D2-945F-00C04FB984F9}'
        elif source == 'dc':
            queried_gponame = '{6AC1786C-016F-11D2-945F-00C04FB984F9}'
        gpo = self.get_netgpo(queried_domain=queried_domain, queried_gponame=queried_gponame)[0]

        gpttmpl_path = '{}\\MACHINE\\Microsoft\\Windows NT\\SecEdit\\GptTmpl.inf'.format(gpo.gpcfilesyspath)
        gpttmpl = self.get_gpttmpl(gpttmpl_path)

        if source == 'domain':
            return gpttmpl
        elif source == 'dc':
            if not resolve_sids:
                return gpttmpl
            else:
                import inspect
                try:
                    privilege_rights_policy = gpttmpl.privilegerights
                except AttributeError:
                    self._logger.critical('Could not parse privilegerights from the DC policy, SIDs will not be resolved')
                    return gpttmpl

                members = inspect.getmembers(privilege_rights_policy, lambda x: not(inspect.isroutine(x)))
                with NetRequester(self._domain_controller, self._domain, self._user,
                                  self._password, self._lmhash, self._nthash, self._do_kerberos, self._do_tls) as net_requester:
                    for attr in privilege_rights_policy._attributes_dict:
                        attribute = privilege_rights_policy._attributes_dict[attr]
                        if not isinstance(attribute, list):
                            sids = [attribute]
                        else:
                            sids = attribute
                        resolved_sids = list()
                        for sid in sids:
                            if not sid:
                                continue
                            sid = sid.replace('*', '')
                            try:
                                resolved_sid = net_requester.get_adobject(queried_sid=sid, queried_domain=self._queried_domain)[0]
                            except IndexError:
                                self._logger.warning('We did not manage to resolve this SID ({}) against the DC'.format(sid))
                                resolved_sid = sid
                            else:
                                resolved_sid = resolved_sid.distinguishedname.split(',')[:2]
                                resolved_sid = resolved_sid[1] + '\\' + resolved_sid[0]
                                resolved_sid = resolved_sid.replace('CN=', '')
                            finally:
                                resolved_sids.append(resolved_sid)
                        if len(resolved_sids) == 1:
                            resolved_sids = resolved_sids[0]
                        privilege_rights_policy._attributes_dict[attr] = resolved_sids

                gpttmpl.privilegerights = privilege_rights_policy

                return gpttmpl

    def _get_groupsxml(self, groupsxml_path, gpo_display_name):
        gpo_groups = list()

        content_io = BytesIO()

        groupsxml_path_split = groupsxml_path.split('\\')
        gpo_name = groupsxml_path_split[6]
        target = self._domain_controller
        share = groupsxml_path_split[3]
        file_name = '\\'.join(groupsxml_path_split[4:])

        smb_connection = SMBConnection(remoteName=target, remoteHost=target)
        if self._do_kerberos:
            smb_connection.kerberosLogin(self._user, self._password, self._domain,
                                 self._lmhash, self._nthash)
        else:
            smb_connection.login(self._user, self._password, self._domain,
                                 self._lmhash, self._nthash)

        self._logger.debug('Get File: Share = {0}, file_name ={1}'.format(share, file_name))
        smb_connection.connectTree(share)
        try:
            smb_connection.getFile(share, file_name, content_io.write)
        except SessionError:
            self._logger.warning('Error while getting the file {}, skipping...'.format(file_name))
            return list()

        content = content_io.getvalue().replace(b'\r', b'')
        groupsxml_soup = BeautifulSoup(content.decode('utf-8'), 'xml')
        for group in groupsxml_soup.find_all('Group'):
            members = list()
            memberof = list()

            raw_xml_member = group.Properties.find_all('Member')
            if not raw_xml_member:
                continue

            local_sid = group.Properties.get('groupSid', str())

            if not local_sid:
                if 'administrators' in group.Properties['groupName'].lower():
                    local_sid = 'S-1-5-32-544'
                elif 'remote desktop' in group.Properties['groupName'].lower():
                    local_sid = 'S-1-5-32-555'
                else:
                    local_sid = group.Properties['groupName']
            memberof.append(local_sid)

            for member in raw_xml_member:
                if not member['action'].lower() == 'add':
                    continue
                if member['sid']:
                    members.append(member['sid'])
                else:
                    members.append(member['name'])

            if members or memberof:
                # TODO: implement filter support (seems like a pain in the ass,
                # I'll do it if the feature is asked). PowerView also seems to
                # have the barest support for filters, so ¯\_(ツ)_/¯

                gpo_group = GPOGroup(list())
                gpo_group._attributes_dict['gpodisplayname'] = gpo_display_name
                gpo_group._attributes_dict['gponame'] = gpo_name
                gpo_group._attributes_dict['gpopath'] = groupsxml_path
                gpo_group._attributes_dict['members'] = members
                gpo_group._attributes_dict['memberof'] = memberof

                gpo_groups.append(gpo_group)

        return gpo_groups

    def _get_groupsgpttmpl(self, gpttmpl_path, gpo_display_name):
        gpo_groups = list()

        gpt_tmpl = self.get_gpttmpl(gpttmpl_path)
        gpo_name = gpttmpl_path.split('\\')[6]

        try:
            group_membership = gpt_tmpl.groupmembership
        except AttributeError:
            return list()

        membership = group_membership._attributes_dict

        for ma,mv in membership.items():
            if not mv:
                continue
            members = list()
            memberof = list()
            if ma.lower().endswith('__memberof'):
                members.append(ma.upper().lstrip('*').replace('__MEMBEROF', ''))
                if not isinstance(mv, list):
                    memberof_list = [mv]
                else:
                    memberof_list = mv
                memberof += [x.lstrip('*') for x in memberof_list]
            elif ma.lower().endswith('__members'):
                memberof.append(ma.upper().lstrip('*').replace('__MEMBERS', ''))
                if not isinstance(mv, list):
                    members_list = [mv]
                else:
                    members_list = mv
                members += [x.lstrip('*') for x in members_list]

            if members and memberof:
                gpo_group = GPOGroup(list())
                gpo_group.add_attributes({'gpodisplayname' : gpo_display_name})
                gpo_group.add_attributes({'gponame' : gpo_name})
                gpo_group.add_attributes({'gpopath' : gpttmpl_path})
                gpo_group.add_attributes({'members' : members})
                gpo_group.add_attributes({'memberof' : memberof})

                gpo_groups.append(gpo_group)

        return gpo_groups

    def get_netgpogroup(self, queried_gponame='*', queried_displayname=str(),
                        queried_domain=str(), ads_path=str(), resolve_sids=False):
        results = list()
        gpos = self.get_netgpo(queried_gponame=queried_gponame,
                               queried_displayname=queried_displayname,
                               queried_domain=queried_domain,
                               ads_path=ads_path)

        for gpo in gpos:
            gpo_display_name = gpo.displayname

            groupsxml_path = '{}\\MACHINE\\Preferences\\Groups\\Groups.xml'.format(gpo.gpcfilesyspath)
            gpttmpl_path = '{}\\MACHINE\\Microsoft\\Windows NT\\SecEdit\\GptTmpl.inf'.format(gpo.gpcfilesyspath)

            results += self._get_groupsxml(groupsxml_path, gpo_display_name)
            try:
                results += self._get_groupsgpttmpl(gpttmpl_path, gpo_display_name)
            except SessionError:
                # If the GptTmpl file doesn't exist, we skip this
                self._logger.warning('Error while getting the file {}, skipping...'.format(gpttmpl_path,))
                pass

        if resolve_sids:
            for gpo_group in results:
                members = gpo_group.members
                memberof = gpo_group.memberof

                resolved_members = list()
                resolved_memberof = list()
                with NetRequester(self._domain_controller, self._domain, self._user,
                                  self._password, self._lmhash, self._nthash, self._do_kerberos, self._do_tls) as net_requester:
                    for member in members:
                        try:
                            resolved_member = net_requester.get_adobject(queried_sid=member, queried_domain=self._queried_domain)[0]
                            resolved_member = resolved_member.distinguishedname
                        except IndexError:
                            self._logger.warning('We did not manage to resolve this SID ({}) against the DC'.format(member))
                            resolved_member = member
                        finally:
                            resolved_members.append(resolved_member)
                    gpo_group._attributes_dict['members'] = resolved_members

                    for member in memberof:
                        try:
                            resolved_member = net_requester.get_adobject(queried_sid=member, queried_domain=self._queried_domain)[0]
                            resolved_member = resolved_member.distinguishedname
                        except IndexError:
                            self._logger.warning('We did not manage to resolve this SID ({}) against the DC'.format(member))
                            resolved_member = member
                        finally:
                            resolved_memberof.append(resolved_member)
                    gpo_group._attributes_dict['memberof'] = memberof = resolved_memberof
        return results

    def find_gpocomputeradmin(self, queried_computername=str(),
                                 queried_ouname=str(), queried_domain=str(),
                                 recurse=False):

        results = list()
        if (not queried_computername) and (not queried_ouname):
            raise ValueError('You must specify either a computer name or an OU name')

        net_requester = NetRequester(self._domain_controller, self._domain, self._user,
                                     self._password, self._lmhash, self._nthash, self._do_kerberos,
                                     self._do_tls)
        if queried_computername:
            computers = net_requester.get_netcomputer(queried_computername=queried_computername,
                                                      queried_domain=queried_domain,
                                                      full_data=True)
            if not computers:
                raise ValueError('Computer {} not found'.format(queried_computername))

            target_ous = list()
            for computer in computers:
                dn = computer.distinguishedname
                for x in dn.split(','):
                    if x.startswith('OU='):
                        target_ous.append(dn[dn.find(x):])
        else:
            target_ous = [queried_ouname]

        # Hack to save the base DN for later
        if not queried_domain:
            saved_queried_domain = ','.join('dc={}'.format(x) for x in net_requester._domain.split('.'))[3:]

        gpo_groups = list()
        for target_ou in target_ous:
            ous = net_requester.get_netou(ads_path=target_ou, queried_domain=queried_domain,
                                          full_data=True)
            for ou in ous:
                try:
                    gplinks = ou.gplink.strip('[]').split('][')
                except AttributeError:
                    continue
                for gplink in gplinks:
                    gplink = gplink.split(';')[0]
                    gpo_groups = self.get_netgpogroup(queried_domain=queried_domain,
                                                      ads_path=gplink)
                    for gpo_group in gpo_groups:
                        for member in gpo_group.members:
                            obj = net_requester.get_adobject(queried_sid=member,
                                                             queried_domain=saved_queried_domain)[0]
                            gpo_computer_admin = GPOComputerAdmin(list())
                            gpo_computer_admin.add_attributes({'computername' : queried_computername})
                            gpo_computer_admin.add_attributes({'ou' : target_ou})
                            gpo_computer_admin.add_attributes({'gpodisplayname' : gpo_group.gpodisplayname})
                            gpo_computer_admin.add_attributes({'gpopath' : gpo_group.gpopath})
                            gpo_computer_admin.add_attributes({'objectname' : obj.name})
                            gpo_computer_admin.add_attributes({'objectdn' : obj.distinguishedname})
                            gpo_computer_admin.add_attributes({'objectsid' : obj.objectsid})
                            gpo_computer_admin.add_attributes({'isgroup' : (obj.samaccounttype == 'GROUP_OBJECT')})

                            results.append(gpo_computer_admin)

                            if recurse and gpo_computer_admin.isgroup:
                                groups_to_resolve = [gpo_computer_admin.objectsid]
                                while groups_to_resolve:
                                    group_to_resolve = groups_to_resolve.pop(0)

                                    group_members = net_requester.get_netgroupmember(queried_sid=group_to_resolve,
                                                                                     queried_domain=saved_queried_domain,
                                                                                     full_data=True)
                                    for group_member in group_members:
                                        gpo_computer_admin = GPOComputerAdmin(list())
                                        gpo_computer_admin.add_attributes({'computername' : queried_computername})
                                        gpo_computer_admin.add_attributes({'ou' : target_ou})
                                        gpo_computer_admin.add_attributes({'gpodisplayname' : gpo_group.gpodisplayname})
                                        gpo_computer_admin.add_attributes({'gpopath' : gpo_group.gpopath})
                                        gpo_computer_admin.add_attributes({'objectname' : group_member.samaccountname})
                                        gpo_computer_admin.add_attributes({'objectdn' : group_member.distinguishedname})
                                        gpo_computer_admin.add_attributes({'objectsid' : group_member.objectsid})
                                        gpo_computer_admin.add_attributes({'isgroup' : (group_member.samaccounttype == 'GROUP_OBJECT')})

                                        results.append(gpo_computer_admin)

                                        if gpo_computer_admin.isgroup:
                                            groups_to_resolve.append(group_member.objectsid)

        return results

    def find_gpolocation(self, queried_username=str(), queried_groupname=str(),
                         queried_localgroup=str(), queried_domain=str()):
        results = list()
        net_requester = NetRequester(self._domain_controller, self._domain, self._user,
                                     self._password, self._lmhash, self._nthash, self._do_kerberos,
                                     self._do_tls)
        if queried_username:
                try:
                    user = net_requester.get_netuser(queried_username=queried_username,
                                                     queried_domain=self._queried_domain)[0]
                except IndexError:
                    raise ValueError('Username \'{}\' was not found'.format(queried_username))
                else:
                    target_sid = [user.objectsid]
                    object_sam_account_name = user.samaccountname
                    object_distinguished_name = user.distinguishedname
        elif queried_groupname:
                try:
                    group = net_requester.get_netgroup(queried_groupname=queried_groupname,
                                                       queried_domain=self._queried_domain,
                                                       full_data=True)[0]
                except IndexError:
                    raise ValueError('Group name \'{}\' was not found'.format(queried_groupname))
                else:
                    target_sid = [group.objectsid]
                    object_sam_account_name = group.samaccountname
                    object_distinguished_name = group.distinguishedname
        else:
            raise ValueError('You must specify either a username or a group name')

        if 'admin' in queried_localgroup.lower():
            local_sid = 'S-1-5-32-544'
        elif 'rdp' in queried_localgroup.lower():
            local_sid = 'S-1-5-32-555'
        elif queried_localgroup.upper().startswith('S-1-5'):
            local_sid = queried_localgroup
        else:
            raise ValueError('The queried local group must be in \'Administrators\', ' \
                    '\'RDP\', or a \'S-1-5\' type SID')

        object_groups = net_requester.get_netgroup(queried_username=object_sam_account_name,
                                                   queried_domain=queried_domain)
        for object_group in object_groups:
            try:
                object_group_sid = net_requester.get_adobject(queried_sam_account_name=object_group.samaccountname,
                                                              queried_domain=self._queried_domain)[0].objectsid
            except IndexError:
                # We may have the name of the group, but not its sam account name
                self._logger.warning('We may have the name of the group, but not its sam account name.')
                try:
                    object_group_sid = net_requester.get_adobject(queried_name=object_group.samaccountname,
                                                                  queried_domain=self._queried_domain)[0].objectsid
                except IndexError:
                    # Freak accident when someone is a member of a group, but
                    # we can't find the group in the AD
                    self._logger.warning('Freak accident when someone is a member of a group, but we can\'t find the group in the AD,'
                                         'see DEBUG level for more info')
                    self._logger.debug('Dumping the mysterious object = {}'.format(object_group))
                    continue

            target_sid.append(object_group_sid)

        gpo_groups = list()
        for gpo_group in self.get_netgpogroup(queried_domain=queried_domain):
            try:
                for member in gpo_group.members:
                    member = member
                    if not member.upper().startswith('S-1-5'):
                        try:
                            member = net_requester.get_adobject(queried_sam_account_name=member,
                                                                queried_domain=self._queried_domain)[0].objectsid
                        except (IndexError, AttributeError):
                            continue
                    if (member.upper() in target_sid) or (member.lower() in target_sid):
                        if (local_sid.upper() in gpo_group.memberof) or \
                                (local_sid.lower() in gpo_group.memberof):
                            gpo_groups.append(gpo_group)
                            break
            except AttributeError:
                continue

        for gpo_group in gpo_groups:
            gpo_guid = gpo_group.gponame
            ous = net_requester.get_netou(queried_domain=self._queried_domain,
                                          queried_guid=gpo_guid, full_data=True)
            for ou in ous:
                ou_distinguishedname = 'LDAP://{}'.format(ou.distinguishedname)
                # TODO: support filters for GPO
                ou_computers = [x.dnshostname for x in \
                        net_requester.get_netcomputer(queried_domain=self._queried_domain,
                                                      ads_path=ou_distinguishedname)]
                gpo_location = GPOLocation(list())
                gpo_location.add_attributes({'objectname' : object_distinguished_name})
                gpo_location.add_attributes({'gponame' : gpo_group.gpodisplayname})
                gpo_location.add_attributes({'gpoguid' : gpo_guid})
                gpo_location.add_attributes({'containername' : ou.distinguishedname})
                gpo_location.add_attributes({'computers' : ou_computers})

                results.append(gpo_location)

        return results




# This file is part of PywerView.

# PywerView is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PywerView is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PywerView.  If not, see <http://www.gnu.org/licenses/>.

# Yannick Méheut [yannick (at) meheut (dot) org] - Copyright © 2023

import logging
import inspect

class RPCObject:
    def __init__(self, obj):
        logger = logging.getLogger('pywerview_main_logger.RPCObject')
        logger.ULTRA = 5
        self._logger = logger

        attributes = dict()
        try:
            for key in obj.fields.keys():
                attributes[key] = obj[key]
        except AttributeError:
            attributes = obj
        self.add_attributes(attributes)

    def add_attributes(self, attributes):
        self._logger.log(self._logger.ULTRA,'RPCObject instancied with the following attributes : {}'.format(attributes))
        for key, value in attributes.items():
            key = key.lower()
            if key in ('wkui1_logon_domain', 'wkui1_logon_server',
                       'wkui1_oth_domains', 'wkui1_username',
                       'sesi10_cname', 'sesi10_username'):
                value = value.rstrip('\x00')

            setattr(self, key.lower(), value)

    def __str__(self):
        s = str()
        members = inspect.getmembers(self, lambda x: not(inspect.isroutine(x)))
        max_length = 0
        #TODO: we redefine here because of a bug when used with invoke-processhunter
        self._logger.ULTRA = 5
        for member in members:
            if not member[0].startswith('_'):
                if len(member[0]) > max_length:
                    max_length = len(member[0])
        for member in members:
            self._logger.log(self._logger.ULTRA,'Trying to print : attribute name = {0} / value = {1}'.format(member[0], member[1]))
            if not member[0].startswith('_'):
                s += '{}: {}{}\n'.format(member[0], ' ' * (max_length - len(member[0])), member[1])

        s = s[:-1]
        return s

    def __repr__(self):
        return str(self)

    def to_json(self):
        members = inspect.getmembers(self, lambda x: not(inspect.isroutine(x)))
        results = dict()
        for member in members:
            if not member[0].startswith('_'):
                results[member[0]]=member[1]
        return(results)

class TargetUser(RPCObject):
    pass

class Session(RPCObject):
    pass

class Share(RPCObject):
    pass

class WkstaUser(RPCObject):
    pass

class Group(RPCObject):
    pass

class Disk(RPCObject):
    def __init__(self, obj):
        RPCObject.__init__(self, obj)
        self.disk = self.disk.rstrip("\x00")

class Process(RPCObject):
    def __init__(self, obj):
        RPCObject.__init__(self, obj)
        self.user = str(self.user)
        self.domain = str(self.domain)

class Event(RPCObject):
    pass




# This file is part of PywerView.

# PywerView is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PywerView is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PywerView.  If not, see <http://www.gnu.org/licenses/>.

# Yannick Méheut [yannick (at) meheut (dot) org] - Copyright © 2023

import random
import multiprocessing
import select

import pywerview.objects.rpcobjects as rpcobj
from pywerview.functions.net import NetRequester
from pywerview.functions.misc import Misc
from pywerview.worker.hunting import UserHunterWorker, ProcessHunterWorker, EventHunterWorker

class Hunter(NetRequester):
    def __init__(self, target_computer, domain=str(), user=(), password=str(),
                 lmhash=str(), nthash=str(), do_kerberos=False, do_tls=False,
                 domain_controller=str(), queried_domain=str()):
        NetRequester.__init__(self, target_computer, domain, user, password,
                              lmhash, nthash, do_kerberos, do_tls, domain_controller)
        self._target_domains = list()
        self._target_computers = list()
        self._target_users = list()
        self._parent_pipes = list()
        self._workers = list()

    def _build_target_domains(self, queried_domain=str()):
        # TODO: implement forest search
        if not queried_domain:
            queried_domain = self._domain
        self._queried_domain = queried_domain
        self._target_domains = [self._queried_domain]

    def _build_target_computers(self, queried_computername=list(), queried_computerfile=None,
                                queried_computerfilter=str(), queried_computeradspath=str(),
                                unconstrained=False, stealth=False,
                                stealth_source=['dfs', 'dc', 'file']):
        if queried_computername:
            self._target_computers = queried_computername

        if not self._target_computers:
            if queried_computerfile:
                self._logger.debug('Computer file provided: {}'.format(queried_computerfile.name))
                with queried_computerfile as _:
                    self._target_computers = [x.rstrip('\n') for x in queried_computerfile.readlines()]

            elif stealth:
                self._logger.debug('Stealth mode enabled')
                for target_domain in self._target_domains:
                    for source in stealth_source:
                        if source == 'dfs':
                            self._target_computers += [x.remoteservername \
                                    for x in self.get_dfsshare(queried_domain=target_domain)]
                        elif source == 'dc':
                            self._target_computers += [x.dnshostname \
                                    for x in self.get_netdomaincontroller(queried_domain=target_domain)]
                        elif source == 'file':
                            self._target_computers += [x.dnshostname \
                                    for x in self.get_netfileserver(queried_domain=target_domain)]
            else:
                self._logger.debug('Computer list not provided by the user, we retrieve it with get-netcomputer')
                for target_domain in self._target_domains:
                    self._target_computers = [x.dnshostname for x in self.get_netcomputer(
                        queried_domain=target_domain, unconstrained=unconstrained,
                        ads_path=queried_computeradspath, custom_filter=queried_computerfilter)]

        # Hack to remove empty arrays from the list
        self._target_computers = [x for x in self._target_computers if x]

        # TODO: automatically convert server names to IP address (DNS, LLMNR, NBT-NS, etc.)
        self._target_computers = list(set(self._target_computers))
        random.shuffle(self._target_computers)

        self._logger.debug('{} computers in the target list'.format(len(self._target_computers)))
        self._logger.log(self._logger.ULTRA,'Target computers: {}'.format(self._target_computers))

        if not self._target_computer:
            raise ValueError('No computers to search against')

    def _build_target_users(self, queried_groupname=str(), target_server=str(),
                            queried_username=str(), queried_userfilter=str(),
                            queried_useradspath=str(), queried_userfile=None,
                            admin_count=False, allow_delegation=False,
                            show_all=False, foreign_users=False):
        if show_all or foreign_users:
            attributes = {'memberdomain': str(), 'membername': str()}
            self._target_users.append(rpcobj.TargetUser(attributes))
        elif target_server:
            self._logger.debug('Target server: {}, we hunt for its localadmins'.format(target_server))
            with NetRequester(target_server, domain, user, password, lmhash,
                              nthash, do_kerberos, do_tls, domain_controller) as target_server_requester:
                for x in target_server_requester.get_netlocalgroup(recurse=True):
                    if x.isdomain and not x.isgroup:
                        attributes = {'memberdomain': x.name.split('/')[0].lower(),
                                      'membername': x.name.split('/')[1].lower()}

                        self._target_users.append(rpcobj.TargetUser(attributes))
        elif queried_userfile:
            self._logger.debug('User file provided: {}'.format(queried_userfile.name))
            with queried_userfile as _:
                for x in queried_userfile.readlines():
                    attributes = dict()
                    attributes['membername'] = x.rstrip('\n')
                    attributes['memberdomain'] = self._target_domains[0]

                    self._target_users.append(rpcobj.TargetUser(attributes))
        elif queried_username:
            self._logger.debug('Queried user: {}'.format(queried_username))
            attributes = dict()
            attributes['membername'] = queried_username.lower()
            attributes['memberdomain'] = self._target_domains[0]

            self._target_users.append(rpcobj.TargetUser(attributes))
        elif queried_useradspath or queried_userfilter or admin_count or allow_delegation:
            self._logger.debug('Custom query used, launching get-netuser')
            for target_domain in self._target_domains:
                for x in self.get_netuser(ads_path=queried_useradspath,
                                          custom_filter=queried_userfilter,
                                          admin_count=admin_count,
                                          allow_delegation=allow_delegation,
                                          queried_domain=target_domain):
                            attributes = dict()
                            attributes['memberdomain'] = target_domain
                            attributes['membername'] = x.samaccountname

                            self._target_users.append(rpcobj.TargetUser(attributes))
        else:
            for target_domain in self._target_domains:
                self._target_users += self.get_netgroupmember(queried_domain=target_domain,
                                                              queried_groupname=queried_groupname,
                                                              recurse=True)

        # Hack to remove empty arrays from the list
        self._target_users = [x for x in self._target_users if x]
        self._target_users = list(set(self._target_users))

        self._logger.debug('{} users in the target list'.format(len(self._target_users)))
        self._logger.log(self._logger.ULTRA,'Target users: {}'.format(self._target_users))

        if (not show_all) and (not foreign_users) and (not self._target_users):
            raise ValueError('No users to search for')

    def _build_workers(self, threads, worker_class, worker_args):
        for i in range(threads):
            parent_pipe, worker_pipe = multiprocessing.Pipe()
            self._parent_pipes.append(parent_pipe)
            worker = worker_class(worker_pipe, self._domain, self._user,
                                            self._password, self._lmhash, self._nthash,
                                            self._do_kerberos, self._do_tls, *worker_args)

            worker.start()
            self._workers.append(worker)

    def _process_workers(self):
        jobs_done, total_jobs = 0, len(self._target_computers)
        try:
            while jobs_done < total_jobs:
                if self._target_computers:
                    write_watch_list = self._parent_pipes
                else:
                    write_watch_list = list()
                rlist, wlist, _ = select.select(self._parent_pipes, write_watch_list, list())

                for readable in rlist:
                    jobs_done += 1
                    results = readable.recv()
                    for result in results:
                        yield result
                for writable in wlist:
                    try:
                        target_computer = self._target_computers.pop(0)
                        writable.send(target_computer)
                    except IndexError:
                        pass
        except KeyboardInterrupt:
            pass
        finally:
            for worker in self._workers:
                worker.terminate()

class UserHunter(Hunter):
    def invoke_userhunter(self, queried_computername=list(), queried_computerfile=None,
            queried_computerfilter=str(), queried_computeradspath=str(),
            unconstrained=False, queried_groupname=str(), target_server=str(),
            queried_username=str(), queried_userfilter=str(), queried_useradspath=str(),
            queried_userfile=None, threads=1, admin_count=False, allow_delegation=False,
            stop_on_success=False, check_access=False, queried_domain=str(), stealth=False,
            stealth_source=['dfs', 'dc', 'file'], show_all=False, foreign_users=False):

        self._logger.debug('Building target domain')

        self._build_target_domains(queried_domain)

        self._logger.debug('Building target computers')

        self._build_target_computers(queried_computername=queried_computername,
                                     queried_computerfile=queried_computerfile,
                                     queried_computerfilter=queried_computerfilter,
                                     queried_computeradspath=queried_computeradspath,
                                     unconstrained=unconstrained, stealth=stealth,
                                     stealth_source=stealth_source)

        self._logger.debug('Building target users')

        self._build_target_users(queried_groupname=queried_groupname,
                                 target_server=target_server,
                                 queried_username=queried_username,
                                 queried_userfilter=queried_userfilter,
                                 queried_useradspath=queried_useradspath,
                                 queried_userfile=queried_userfile,
                                 admin_count=admin_count, allow_delegation=allow_delegation,
                                 show_all=show_all, foreign_users=foreign_users)

        if foreign_users:
            with Misc(self._domain_controller, self._domain, self._user,
                      self._password, self._lmhash, self._nthash, self._do_kerberos, self._do_tls) as misc_requester:
                domain_sid = misc_requester.get_domainsid(queried_domain)
                domain_short_name = misc_requester.convert_sidtont4(domain_sid).split('\\')[0]
        else:
            domain_short_name = None

        self._logger.debug('Launching workers ({} threads)...'.format(threads))

        self._build_workers(threads, UserHunterWorker, (foreign_users, stealth,
                                                        self._target_users,
                                                        domain_short_name, check_access))
        return self._process_workers()

class ProcessHunter(Hunter):
    def invoke_processhunter(self, queried_computername=list(), queried_computerfile=None,
            queried_computerfilter=str(), queried_computeradspath=str(),
            queried_processname=list(), queried_groupname=str(), target_server=str(),
            queried_username=str(), queried_userfilter=str(), queried_useradspath=str(),
            queried_userfile=None, threads=1, stop_on_success=False, queried_domain=str(),
            show_all=False):

        self._logger.debug('Building target domain')

        self._build_target_domains(queried_domain)

        self._logger.debug('Building target computers')

        self._build_target_computers(queried_computername=queried_computername,
                                     queried_computerfile=queried_computerfile,
                                     queried_computerfilter=queried_computerfilter,
                                     queried_computeradspath=queried_computeradspath)

        self._logger.debug('Building target users')

        self._build_target_users(queried_groupname=queried_groupname,
                                 target_server=target_server,
                                 queried_username=queried_username,
                                 queried_userfilter=queried_userfilter,
                                 queried_useradspath=queried_useradspath,
                                 queried_userfile=queried_userfile,
                                 show_all=show_all)

        self._logger.debug('Launching workers ({} threads)...'.format(threads))

        self._build_workers(threads, ProcessHunterWorker, (queried_processname,
                                                           self._target_users))

        return self._process_workers()

class EventHunter(Hunter):
    def invoke_eventhunter(self, queried_computername=list(), queried_computerfile=None,
                           queried_computerfilter=str(), queried_computeradspath=str(),
                           queried_groupname=str(), target_server=str(), queried_username=str(),
                           queried_useradspath=str(), queried_userfilter=str(),
                           queried_userfile=None, threads=1, queried_domain=str(),
                           search_days=3):

        self._logger.debug('Building target domain')

        self._build_target_domains(queried_domain)

        self._logger.debug('Building target computers')

        self._build_target_computers(queried_computername=queried_computername,
                                     queried_computerfile=queried_computerfile,
                                     queried_computerfilter=queried_computerfilter,
                                     queried_computeradspath=queried_computeradspath)

        self._logger.debug('Building target users')

        self._build_target_users(queried_groupname=queried_groupname,
                                 target_server=target_server,
                                 queried_username=queried_username,
                                 queried_userfilter=queried_userfilter,
                                 queried_useradspath=queried_useradspath,
                                 queried_userfile=queried_userfile)

        self._logger.debug('Launching workers ({} threads)...'.format(threads))

        self._build_workers(threads, EventHunterWorker, (search_days,
                                                         self._target_users))

        return self._process_workers()



# This file is part of PywerView.

# PywerView is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PywerView is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PywerView.  If not, see <http://www.gnu.org/licenses/>.

# Yannick Méheut [yannick (at) meheut (dot) org] - Copyright © 2023

from impacket.dcerpc.v5.rpcrt import DCERPCException
from impacket.dcerpc.v5 import scmr, drsuapi

from pywerview.requester import LDAPRPCRequester
import pywerview.functions.net

class Misc(LDAPRPCRequester):
    @LDAPRPCRequester._rpc_connection_init(r'\drsuapi')
    def convert_sidtont4(self, sid):

        # We get a DRS handle, shamelessly stolen from secretsdump.py
        request = drsuapi.DRSBind()
        request['puuidClientDsa'] = drsuapi.NTDSAPI_CLIENT_GUID
        drs = drsuapi.DRS_EXTENSIONS_INT()
        drs['cb'] = len(drs) #- 4
        drs['dwFlags'] = drsuapi.DRS_EXT_GETCHGREQ_V6 | drsuapi.DRS_EXT_GETCHGREPLY_V6 | drsuapi.DRS_EXT_GETCHGREQ_V8 | \
                         drsuapi.DRS_EXT_STRONG_ENCRYPTION
        drs['SiteObjGuid'] = drsuapi.NULLGUID
        drs['Pid'] = 0
        drs['dwReplEpoch'] = 0
        drs['dwFlagsExt'] = 0
        drs['ConfigObjGUID'] = drsuapi.NULLGUID
        drs['dwExtCaps'] = 0xffffffff
        request['pextClient']['cb'] = len(drs)
        request['pextClient']['rgb'] = list(str(drs))

        hdrs = self._rpc_connection.request(request)['phDrs']

        resp = drsuapi.hDRSCrackNames(self._rpc_connection, hdrs, 0x0, 11, 2, (sid,))

        return resp['pmsgOut']['V1']['pResult']['rItems'][0]['pName']

    def get_domainsid(self, queried_domain=str()):

        with pywerview.functions.net.NetRequester(self._domain_controller, self._domain, self._user,
                                                  self._password, self._lmhash, self._nthash,
                                                  self._do_kerberos, self._do_tls,
                                                  self._user_cert, self._user_key) as r:
            domain_controllers = r.get_netdomaincontroller(queried_domain=queried_domain)

        if domain_controllers:
            primary_dc = domain_controllers[0]
            domain_sid = primary_dc.objectsid

            # we need to retrieve the domain sid from the controller sid
            domain_sid = '-'.join(domain_sid.split('-')[:-1])
        else:
            domain_sid = None

        return domain_sid

    @LDAPRPCRequester._rpc_connection_init(r'\svcctl')
    def invoke_checklocaladminaccess(self):

        try:
            # 0xF003F - SC_MANAGER_ALL_ACCESS
            # http://msdn.microsoft.com/en-us/library/windows/desktop/ms685981(v=vs.85).aspx
            ans = scmr.hROpenSCManagerW(self._rpc_connection,
                                        '{}\x00'.format(self._target_computer),
                                        'ServicesActive\x00', 0xF003F)
        except DCERPCException:
            return False

        return True




#!/usr/bin/env python
#
# This file is part of PywerView.

# PywerView is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PywerView is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PywerView.  If not, see <http://www.gnu.org/licenses/>.

# Yannick Méheut [yannick (at) meheut (dot) org] - Copyright © 2023

from pywerview.functions.net import NetRequester
from pywerview.functions.gpo import GPORequester
from pywerview.functions.misc import Misc
from pywerview.functions.hunting import UserHunter, ProcessHunter, EventHunter

def get_adobject(domain_controller, domain, user, password=str(),
                lmhash=str(), nthash=str(), do_simple=False, do_kerberos=False,
                do_tls=False, user_cert=str(), user_key=str(),
                queried_domain=str(), queried_sid=str(), queried_name=str(),
                queried_sam_account_name=str(), ads_path=str(), attributes=list(),
                custom_filter=str()):
    requester = NetRequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_simple, do_kerberos, do_tls,
                                 user_cert, user_key)
    return requester.get_adobject(queried_domain=queried_domain,
                    queried_sid=queried_sid, queried_name=queried_name,
                    queried_sam_account_name=queried_sam_account_name,
                    ads_path=ads_path, attributes=attributes, custom_filter=custom_filter)

def get_objectowner(domain_controller, domain, user, password=str(),
                lmhash=str(), nthash=str(), do_simple=False, do_kerberos=False,
                do_tls=False, user_cert=str(), user_key=str(),
                queried_domain=str(), queried_sid=str(), queried_name=str(),
                queried_sam_account_name=str(), ads_path=str(),
                custom_filter=str(), resolve_sids=False):
    requester = NetRequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_simple, do_kerberos,
                                 do_tls, user_cert, user_key)
    return requester.get_objectowner(queried_domain=queried_domain,
                    queried_sid=queried_sid, queried_name=queried_name,
                    queried_sam_account_name=queried_sam_account_name,
                    ads_path=ads_path, resolve_sids=resolve_sids, custom_filter=custom_filter)

def get_netpki(domain_controller, domain, user, password=str(),
               lmhash=str(), nthash=str(), do_simple=False, do_kerberos=False,
               do_tls=False, user_cert=str(), user_key=str(), queried_domain=str(),
               queried_ca_name=str(), resolve_sids=False, full_data=False):
    requester = NetRequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_simple,do_kerberos,
                                 do_tls, user_cert, user_key)
    return requester.get_netpki(queried_domain=queried_domain,
                    queried_ca_name=queried_ca_name, resolve_sids=resolve_sids, full_data=full_data)

def get_netcerttmpl(domain_controller, domain, user, password=str(),
                    lmhash=str(), nthash=str(), do_simple=False, do_kerberos=False,
                    do_tls=False, user_cert=str(), user_key=str(), queried_domain=str(),
                    resolve_sids=False, full_data=False, queried_ca_name=str()):
    requester = NetRequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_simple, do_kerberos,
                                 do_tls, user_cert, user_key)
    return requester.get_netcerttmpl(queried_domain=queried_domain,
                    resolve_sids=resolve_sids, full_data=full_data, queried_ca_name=queried_ca_name)

def get_netgmsa(domain_controller, domain, user, password=str(),
                lmhash=str(), nthash=str(), do_simple=False, do_kerberos=False,
                do_tls=False, user_cert=str(), user_key=str(),
                queried_domain=str(), queried_sid=str(), queried_name=str(),
                queried_sam_account_name=str(), ads_path=str(), resolve_sids=False):
    requester = NetRequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_simple, do_kerberos,
                                 do_tls, user_cert, user_key)
    return requester.get_netgmsa(queried_domain=queried_domain,
                    queried_sid=queried_sid, queried_name=queried_name,
                    queried_sam_account_name=queried_sam_account_name,
                    ads_path=ads_path, resolve_sids=resolve_sids)

def get_netsmsa(domain_controller, domain, user, password=str(),
                lmhash=str(), nthash=str(), do_simple=False, do_kerberos=False,
                do_tls=False, user_cert=str(), user_key=str(),
                queried_domain=str(), queried_sid=str(), queried_name=str(),
                queried_sam_account_name=str(), ads_path=str()):
    requester = NetRequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_simple, do_kerberos,
                                 do_tls, user_cert, user_key)
    return requester.get_netsmsa(queried_domain=queried_domain,
                    queried_sid=queried_sid, queried_name=queried_name,
                    queried_sam_account_name=queried_sam_account_name,
                    ads_path=ads_path)

def get_objectacl(domain_controller, domain, user, password=str(),
                lmhash=str(), nthash=str(), do_simple=False, do_kerberos=False,
                do_tls=False, queried_domain=str(), queried_sid=str(), queried_name=str(),
                user_cert=str(), user_key=str(),
                queried_sam_account_name=str(), ads_path=str(), sacl=False,
                rights_filter=str(), resolve_sids=False, resolve_guids=False,
                custom_filter=str()):
    requester = NetRequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_simple, do_kerberos,
                                 do_tls, user_cert, user_key)
    return requester.get_objectacl(queried_domain=queried_domain,
                    queried_sid=queried_sid, queried_name=queried_name,
                    queried_sam_account_name=queried_sam_account_name,
                    ads_path=ads_path, sacl=sacl, rights_filter=rights_filter,
                    resolve_sids=resolve_sids, resolve_guids=resolve_guids,
                    custom_filter=custom_filter)

def get_netuser(domain_controller, domain, user, password=str(), lmhash=str(),
                nthash=str(), do_simple=False, do_kerberos=False,
                user_cert=str(), user_key=str(), do_tls=False, queried_username=str(),
                queried_domain=str(), ads_path=str(), admin_count=False, spn=False,
                unconstrained=False, allow_delegation=False, preauth_notreq=False,
                custom_filter=str(), attributes=[]):
    requester = NetRequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_simple, do_kerberos,
                                 do_tls, user_cert, user_key)
    return requester.get_netuser(queried_username=queried_username,
                                    queried_domain=queried_domain, ads_path=ads_path, admin_count=admin_count,
                                    spn=spn, unconstrained=unconstrained, allow_delegation=allow_delegation,
                                    preauth_notreq=preauth_notreq, custom_filter=custom_filter,
                                    attributes=attributes)

def get_netgroup(domain_controller, domain, user, password=str(),
                lmhash=str(), nthash=str(), do_simple=False,
                do_kerberos=False, do_tls=False, user_cert=str(), user_key=str(),
                queried_groupname='*', queried_sid=str(), queried_username=str(),
                queried_domain=str(), ads_path=str(), admin_count=False,
                full_data=False, custom_filter=str()):
    requester = NetRequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_simple, do_kerberos,
                                 do_tls, user_cert, user_key)
    return requester.get_netgroup(queried_groupname=queried_groupname,
                                    queried_sid=queried_sid, queried_username=queried_username,
                                    queried_domain=queried_domain, ads_path=ads_path, admin_count=admin_count,
                                    full_data=full_data, custom_filter=custom_filter)

def get_netcomputer(domain_controller, domain, user, password=str(),
                    lmhash=str(), nthash=str(), do_simple=False,
                    do_kerberos=False, do_tls=False, user_cert=str(), user_key=str(),
                    queried_computername=str(), queried_spn=str(), queried_os=str(),
                    queried_sp=str(), queried_domain=str(), ads_path=str(),
                    printers=False, unconstrained=False, laps_passwords=False, pre_created=False,
                    ping=False, full_data=False, custom_filter=str(), attributes=[]):
    requester = NetRequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_simple, do_kerberos,
                                 do_tls, user_cert, user_key)
    return requester.get_netcomputer(queried_computername=queried_computername,
                                        queried_spn=queried_spn, queried_os=queried_os, queried_sp=queried_sp,
                                        queried_domain=queried_domain, ads_path=ads_path, printers=printers,
                                        unconstrained=unconstrained, laps_passwords=laps_passwords,
                                        pre_created=pre_created, ping=ping, full_data=full_data,
                                        custom_filter=custom_filter, attributes=attributes)

def get_netdomaincontroller(domain_controller, domain, user, password=str(),
                                 lmhash=str(), nthash=str(), do_simple=False,
                                 do_kerberos=False, user_cert=str(), user_key=str(),
                                 do_tls=False, queried_domain=str()):
    requester = NetRequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_simple, do_kerberos,
                                 do_tls, user_cert, user_key)
    return requester.get_netdomaincontroller(queried_domain=queried_domain)

def get_netfileserver(domain_controller, domain, user, password=str(),
                                 lmhash=str(), nthash=str(), do_simple=False,
                                 do_kerberos=False, user_cert=str(), user_key=str(),
                                 do_tls=False, queried_domain=str(), target_users=list()):
    requester = NetRequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_simple, do_kerberos,
                                 do_tls, user_cert, user_key)
    return requester.get_netfileserver(queried_domain=queried_domain,
                                            target_users=target_users)

def get_dfsshare(domain_controller, domain, user, password=str(),
                 lmhash=str(), nthash=str(), do_kerberos=False, do_tls=False,
                 version=['v1', 'v2'], queried_domain=str(), ads_path=str()):
    requester = NetRequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_kerberos, do_tls)
    return requester.get_dfsshare(version=version, queried_domain=queried_domain, ads_path=ads_path)

def get_netou(domain_controller, domain, user, password=str(), lmhash=str(),
              nthash=str(), do_simple=False, do_kerberos=False,
              do_tls=False, queried_domain=str(), queried_ouname='*',
              queried_guid=str(), ads_path=str(), full_data=False,
              user_cert=str(), user_key=str()):
    requester = NetRequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_simple, do_kerberos,
                                 do_tls, user_cert, user_key)
    return requester.get_netou(queried_domain=queried_domain,
                                   queried_ouname=queried_ouname, queried_guid=queried_guid, ads_path=ads_path,
                                   full_data=full_data)

def get_netsite(domain_controller, domain, user, password=str(), lmhash=str(),
                nthash=str(), do_simple=False, do_kerberos=False, do_tls=False,
                queried_domain=str(), queried_sitename=str(), queried_guid=str(),
                ads_path=str(), ads_prefix='CN=Sites,CN=Configuration', full_data=False,
                user_cert=str(), user_key=str()):
    requester = NetRequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_simple, do_kerberos,
                                 do_tls, user_cert, user_key)
    return requester.get_netsite(queried_domain=queried_domain,
                                    queried_sitename=queried_sitename, queried_guid=queried_guid,
                                    ads_path=ads_path, ads_prefix=ads_prefix, full_data=full_data)

def get_netsubnet(domain_controller, domain, user, password=str(),
                  lmhash=str(), nthash=str(), do_simple=False, do_kerberos=False,
                  do_tls=False, queried_domain=str(), queried_sitename=str(),
                  ads_path=str(), ads_prefix='CN=Sites,CN=Configuration',
                  full_data=False, user_cert=str(), user_key=str()):
    requester = NetRequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_simple, do_kerberos,
                                 do_tls, user_cert, user_key)
    return requester.get_netsubnet(queried_domain=queried_domain,
                                       queried_sitename=queried_sitename, ads_path=ads_path, ads_prefix=ads_prefix,
                                       full_data=full_data)

def get_netdomaintrust(domain_controller, domain, user, password=str(),
                  lmhash=str(), nthash=str(), do_simple=False, do_kerberos=False,
                  do_tls=False, queried_domain=str(), user_cert=str(), user_key=str(),
                  full_data=False):
    requester = NetRequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_simple, do_kerberos,
                                 do_tls, user_cert, user_key)
    return requester.get_netdomaintrust(queried_domain=queried_domain, full_data=full_data)

def get_netgroupmember(domain_controller, domain, user, password=str(),
                       lmhash=str(), nthash=str(), do_simple=False, do_kerberos=False,
                       do_tls=False, user_cert=str(), user_key=str(),
                       queried_groupname=str(), queried_sid=str(), queried_domain=str(),
                       ads_path=str(), recurse=False, use_matching_rule=False,
                       full_data=False, custom_filter=str()):
    requester = NetRequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_simple, do_kerberos,
                                 do_tls, user_cert, user_key)
    return requester.get_netgroupmember(queried_groupname=queried_groupname,
                                            queried_sid=queried_sid, queried_domain=queried_domain,
                                            ads_path=ads_path, recurse=recurse,
                                            use_matching_rule=use_matching_rule,
                                            full_data=full_data, custom_filter=custom_filter)

def get_netsession(target_computername, domain, user, password=str(),
                   lmhash=str(), nthash=str(), do_kerberos=False):
    requester = NetRequester(target_computername, domain, user, password,
                                 lmhash, nthash, do_kerberos)
    return requester.get_netsession()

def get_netshare(target_computername, domain, user, password=str(),
                                 lmhash=str(), nthash=str(), do_kerberos=False):
    requester = NetRequester(target_computername, domain, user, password,
                                 lmhash, nthash, do_kerberos)
    return requester.get_netshare()

def get_localdisks(target_computername, domain, user, password=str(),
                                 lmhash=str(), nthash=str(), do_kerberos=False):
    requester = NetRequester(target_computername, domain, user, password,
                                 lmhash, nthash, do_kerberos)
    return requester.get_localdisks()

def get_netdomain(domain_controller, domain, user, password=str(),
                                 lmhash=str(), nthash=str(), do_kerberos=False,
                                 do_tls=False):
    requester = NetRequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_kerberos, do_tls)
    return requester.get_netdomain()

def get_netloggedon(target_computername, domain, user, password=str(),
                                 lmhash=str(), nthash=str(), do_kerberos=False):
    requester = NetRequester(target_computername, domain, user, password,
                                 lmhash, nthash, do_kerberos)
    return requester.get_netloggedon()

def get_netlocalgroup(target_computername, domain_controller, domain, user,
                      password=str(), lmhash=str(), nthash=str(), do_kerberos=False,
                      do_tls=False, queried_groupname=str(), list_groups=False,
                      recurse=False):
    requester = NetRequester(target_computername, domain, user, password,
                                 lmhash, nthash, do_kerberos, do_tls,
                                 domain_controller=domain_controller)
    return requester.get_netlocalgroup(queried_groupname=queried_groupname,
                                           list_groups=list_groups, recurse=recurse)

def get_netprocess(target_computername, domain, user, password=str(),
                   lmhash=str(), nthash=str(), do_kerberos=False):
    requester = NetRequester(target_computername, domain, user, password,
                                 lmhash, nthash, do_kerberos)
    return requester.get_netprocess()

def get_userevent(target_computername, domain, user, password=str(),
                   lmhash=str(), nthash=str(), do_kerberos=False,
                   event_type=['logon', 'tgt'], date_start=5):
    requester = NetRequester(target_computername, domain, user, password,
                                 lmhash, nthash, do_kerberos)
    return requester.get_userevent(event_type=event_type,
                                       date_start=date_start)

def get_netgpo(domain_controller, domain, user, password=str(),
               lmhash=str(), nthash=str(), do_simple=False,
               do_kerberos=False, do_tls=False, user_cert=str(),
               user_key=str(), queried_gponame='*',
               queried_displayname=str(), queried_domain=str(),
               ads_path=str()):
    requester = GPORequester(domain_controller, domain, user, password,
                                lmhash, nthash, do_simple, do_kerberos,
                                do_tls, user_cert, user_key)
    return requester.get_netgpo(queried_gponame=queried_gponame,
                                    queried_displayname=queried_displayname,
                                    queried_domain=queried_domain, ads_path=ads_path)

def get_netpso(domain_controller, domain, user, password=str(),
               lmhash=str(), nthash=str(), do_simple=False,
               do_kerberos=False, do_tls=False, user_cert=str(), user_key=str(),
               queried_psoname='*', queried_displayname=str(), queried_domain=str(),
               ads_path=str()):
    requester = GPORequester(domain_controller, domain, user, password,
                                lmhash, nthash, do_simple, do_kerberos,
                                do_tls, user_cert, user_key)
    return requester.get_netpso(queried_psoname=queried_psoname,
                                    queried_displayname=queried_displayname,
                                    queried_domain=queried_domain, ads_path=ads_path)

def get_domainpolicy(domain_controller, domain, user, password=str(),
                     lmhash=str(), nthash=str(), do_kerberos=False, do_tls=False,
                     source='domain', queried_domain=str(), resolve_sids=False):
    requester = GPORequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_kerberos, do_tls)

    return requester.get_domainpolicy(source=source, queried_domain=queried_domain,
                                          resolve_sids=resolve_sids)

def get_gpttmpl(gpttmpl_path, domain_controller, domain, user, password=str(), lmhash=str(),
                nthash=str(), do_kerberos=False, do_tls=False):
    requester = GPORequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_kerberos, do_tls)

    return requester.get_gpttmpl(gpttmpl_path)

def get_netgpogroup(domain_controller, domain, user, password=str(), lmhash=str(),
                    nthash=str(), do_kerberos=False, do_tls=False, queried_gponame='*',
                    queried_displayname=str(), queried_domain=str(), ads_path=str(),
                    resolve_sids=False):
    requester = GPORequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_kerberos, do_tls)

    return requester.get_netgpogroup(queried_gponame=queried_gponame,
                                         queried_displayname=queried_displayname,
                                         queried_domain=queried_domain,
                                         ads_path=ads_path,
                                         resolve_sids=resolve_sids)

def find_gpocomputeradmin(domain_controller, domain, user, password=str(), lmhash=str(),
                          nthash=str(), do_kerberos=False, do_tls=False, queried_computername=str(),
                          queried_ouname=str(), queried_domain=str(), recurse=False):
    requester = GPORequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_kerberos, do_tls)

    return requester.find_gpocomputeradmin(queried_computername=queried_computername,
                                               queried_ouname=queried_ouname,
                                               queried_domain=queried_domain,
                                               recurse=recurse)

def find_gpolocation(domain_controller, domain, user, password=str(), lmhash=str(),
                     nthash=str(), do_kerberos=False, do_tls=False, queried_username=str(),
                     queried_groupname=str(), queried_localgroup=str(),
                     queried_domain=str()):
    requester = GPORequester(domain_controller, domain, user, password,
                                 lmhash, nthash, do_kerberos, do_tls)
    return requester.find_gpolocation(queried_username=queried_username,
                                          queried_groupname=queried_groupname,
                                          queried_localgroup=queried_localgroup,
                                          queried_domain=queried_domain)

def invoke_checklocaladminaccess(target_computername, domain, user, password=str(),
                                 lmhash=str(), nthash=str(), do_kerberos=False):
    misc = Misc(target_computername, domain, user, password, lmhash, nthash, do_kerberos)

    return misc.invoke_checklocaladminaccess()

def invoke_userhunter(domain_controller, domain, user, password=str(),
                      lmhash=str(), nthash=str(), do_kerberos=False, do_tls=False,
                      queried_computername=list(), queried_computerfile=None,
                      queried_computerfilter=str(), queried_computeradspath=str(),
                      unconstrained=False, queried_groupname=str(), target_server=str(),
                      queried_username=str(), queried_useradspath=str(),
                      queried_userfilter=str(), queried_userfile=None,
                      threads=1, admin_count=False, allow_delegation=False,
                      stop_on_success=False, check_access=False, queried_domain=str(),
                      stealth=False, stealth_source=['dfs', 'dc', 'file'],
                      show_all=False, foreign_users=False):
    user_hunter = UserHunter(domain_controller, domain, user, password,
                             lmhash, nthash, do_kerberos, do_tls)
    return user_hunter.invoke_userhunter(queried_computername=queried_computername,
                                         queried_computerfile=queried_computerfile,
                                         queried_computerfilter=queried_computerfilter,
                                         queried_computeradspath=queried_computeradspath,
                                         unconstrained=unconstrained, queried_groupname=queried_groupname,
                                         target_server=target_server, queried_username=queried_username,
                                         queried_userfilter=queried_userfilter,
                                         queried_useradspath=queried_useradspath, queried_userfile=queried_userfile,
                                         threads=threads, admin_count=admin_count,
                                         allow_delegation=allow_delegation, stop_on_success=stop_on_success,
                                         check_access=check_access, queried_domain=queried_domain, stealth=stealth,
                                         stealth_source=stealth_source, show_all=show_all,
                                         foreign_users=foreign_users)

def invoke_processhunter(domain_controller, domain, user, password=str(),
                         lmhash=str(), nthash=str(), do_kerberos=False, do_tls=False,
                         queried_computername=list(), queried_computerfile=None,
                         queried_computerfilter=str(), queried_computeradspath=str(),
                         queried_processname=list(), queried_groupname=str(), target_server=str(),
                         queried_username=str(), queried_useradspath=str(),
                         queried_userfilter=str(), queried_userfile=None, threads=1,
                         stop_on_success=False, queried_domain=str(), show_all=False):
    process_hunter = ProcessHunter(domain_controller, domain, user, password,
                                   lmhash, nthash, do_kerberos, do_tls)

    return process_hunter.invoke_processhunter(queried_computername=queried_computername,
                                               queried_computerfile=queried_computerfile,
                                               queried_computerfilter=queried_computerfilter,
                                               queried_computeradspath=queried_computeradspath,
                                               queried_processname=queried_processname,
                                               queried_groupname=queried_groupname,
                                               target_server=target_server, queried_username=queried_username,
                                               queried_userfilter=queried_userfilter,
                                               queried_useradspath=queried_useradspath, queried_userfile=queried_userfile,
                                               threads=threads, stop_on_success=stop_on_success,
                                               queried_domain=queried_domain, show_all=show_all)

def invoke_eventhunter(domain_controller, domain, user, password=str(),
                       lmhash=str(), nthash=str(), do_kerberos=False, do_tls=False,
                       queried_computername=list(), queried_computerfile=None,
                       queried_computerfilter=str(), queried_computeradspath=str(),
                       queried_groupname=str(), target_server=str(), queried_username=str(),
                       queried_useradspath=str(), queried_userfilter=str(),
                       queried_userfile=None, threads=1, queried_domain=str(),
                       search_days=3):
    event_hunter = EventHunter(domain_controller, domain, user, password,
                                   lmhash, nthash, do_kerberos, do_tls)

    return event_hunter.invoke_eventhunter(queried_computername=queried_computername,
                                           queried_computerfile=queried_computerfile,
                                           queried_computerfilter=queried_computerfilter,
                                           queried_computeradspath=queried_computeradspath,
                                           queried_groupname=queried_groupname,
                                           target_server=target_server,
                                           queried_userfilter=queried_userfilter,
                                           queried_username=queried_username,
                                           queried_useradspath=queried_useradspath,
                                           queried_userfile=queried_userfile,
                                           search_days=search_days,
                                           threads=threads, queried_domain=queried_domain)




# This file is part of PywerView.

# PywerView is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PywerView is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PywerView.  If not, see <http://www.gnu.org/licenses/>.

# Yannick Méheut [yannick (at) meheut (dot) org] - Copyright © 2023

import logging
import binascii

try:
    from Cryptodome.Hash import MD4
except ImportError:
    from Crypto.Hash import MD4

from impacket.examples.ntlmrelayx.attacks.ldapattack import MSDS_MANAGEDPASSWORD_BLOB
from impacket.ldap.ldaptypes import SR_SECURITY_DESCRIPTOR

__uac_flags = {0x0000001: 'SCRIPT',
               0x0000002: 'ACCOUNTDISABLE',
               0x0000008: 'HOMEDIR_REQUIRED',
               0x0000010: 'LOCKOUT',
               0x0000020: 'PASSWD_NOTREQD',
               0x0000040: 'PASSWD_CANT_CHANGE',
               0x0000080: 'ENCRYPTED_TEXT_PWD_ALLOWED',
               0x0000100: 'TEMP_DUPLICATE_ACCOUNT',
               0x0000200: 'NORMAL_ACCOUNT',
               0x0000800: 'INTERDOMAIN_TRUST_ACCOUNT',
               0x0001000: 'WORKSTATION_TRUST_ACCOUNT',
               0x0002000: 'SERVER_TRUST_ACCOUNT',
               0x0010000: 'DONT_EXPIRE_PASSWORD',
               0x0020000: 'MNS_LOGON_ACCOUNT',
               0x0040000: 'SMARTCARD_REQUIRED',
               0x0080000: 'TRUSTED_FOR_DELEGATION',
               0x0100000: 'NOT_DELEGATED',
               0x0200000: 'USE_DES_KEY_ONLY',
               0x0400000: 'DONT_REQ_PREAUTH',
               0x0800000: 'PASSWORD_EXPIRED',
               0x1000000: 'TRUSTED_TO_AUTH_FOR_DELEGATION',
               0x4000000: 'PARTIAL_SECRETS_ACCOUNT'}

__sat_flags = {0x00000000: 'DOMAIN_OBJECT',
               0x10000000: 'GROUP_OBJECT',
               0x10000001: 'NON_SECURITY_GROUP_OBJECT',
               0x20000000: 'ALIAS_OBJECT',
               0x20000001: 'NON_SECURITY_ALIAS_OBJECT',
               # According to https://docs.microsoft.com/en-us/windows/win32/adschema/a-samaccounttype
               # USER_OBJECT and NORMAL_USER_ACCOUNT are both 0x30000000
               0x30000000: 'USER_OBJECT',
               0x30000001: 'MACHINE_ACCOUNT',
               0x30000002: 'TRUST_ACCOUNT',
               0x40000000: 'APP_BASIC_GROUP',
               0x40000001: 'APP_QUERY_GROUP',
               0x7fffffff: 'ACCOUNT_TYPE_MAX'}

__ace_flags = {0x1: 'object_inherit', 0x2: 'container_inherit',
               0x4: 'non_propagate_inherit', 0x8: 'inherit_only',
               0x10: 'inherited_ace', 0x20: 'audit_successful_accesses',
               0x40: 'audit_failed_access'}

__object_ace_flags = {0x1: 'object_ace_type_present', 0x2: 'inherited_object_ace_type_present'}

# Resources: https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-adts/990fb975-ab31-4bc1-8b75-5da132cd4584
__access_mask = {0x1: 'create_child', 0x2: 'delete_child',
                 0x4: 'list_children', 0x08: 'self',
                 0x10: 'read_property', 0x20: 'write_property',
                 0x40: 'delete_tree', 0x80: 'list_object',
                 0x100: 'extended_right', 0x10000: 'delete',
                 0x20000: 'read_control', 0x40000: 'write_dacl',
                 0x80000: 'write_owner'}

__access_mask_generic = {0xf01ff: 'generic_all', 0x20094: 'generic_read',
                         0x20028: 'generic_write', 0x20004: 'generic_execute'}

# https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-adts/e9a2d23c-c31e-4a6f-88a0-6646fdb51a3c
__trust_attrib = {0x1: 'TRUST_ATTRIBUTE_NON_TRANSITIVE', 
                  0x2: 'TRUST_ATTRIBUTE_UPLEVEL_ONLY',
                  0x4: 'TRUST_ATTRIBUTE_QUARANTINED_DOMAIN', 
                  0x8: 'TRUST_ATTRIBUTE_FOREST_TRANSITIVE',
                  0x10: 'TRUST_ATTRIBUTE_CROSS_ORGANIZATION', 
                  0x20: 'TRUST_ATTRIBUTE_WITHIN_FOREST',
                  0x40: 'TRUST_ATTRIBUTE_TREAT_AS_EXTERNAL',
                  0x80: 'TRUST_ATTRIBUTE_USES_RC4_ENCRYPTION',
                  # TODO: x100 seems not documented ?
                  0x100: 'TRUST_USES_AES_KEYS',
                  0X200: 'TRUST_ATTRIBUTE_CROSS_ORGANIZATION_NO_TGT_DELEGATION',
                  0x400: 'TRUST_ATTRIBUTE_PIM_TRUST',
                  0x800: 'TRUST_ATTRIBUTE_CROSS_ORGANIZATION_ENABLE_TGT_DELEGATION'}

__trust_direction = {0: 'disabled', 1: 'inbound',
                     2: 'outbound', 3: 'bidirectional'}

__trust_type = {1: 'windows_non_active_directory',
                2: 'windows_active_directory', 3: 'mit'}

# https://www.pkisolutions.com/object-identifiers-oid-in-pki/
# sources: Certipy/certipy/lib/constants.py
__ekus = {
    "1.3.6.1.4.1.311.76.6.1": "Windows Update",
    "1.3.6.1.4.1.311.10.3.11": "Key Recovery",
    "1.3.6.1.4.1.311.10.3.25": "Windows Third Party Application Component",
    "1.3.6.1.4.1.311.21.6": "Key Recovery Agent",
    "1.3.6.1.4.1.311.10.3.6": "Windows System Component Verification",
    "1.3.6.1.4.1.311.61.4.1": "Early Launch Antimalware Drive",
    "1.3.6.1.4.1.311.10.3.23": "Windows TCB Component",
    "1.3.6.1.4.1.311.61.1.1": "Kernel Mode Code Signing",
    "1.3.6.1.4.1.311.10.3.26": "Windows Software Extension Verification",
    "2.23.133.8.3": "Attestation Identity Key Certificate",
    "1.3.6.1.4.1.311.76.3.1": "Windows Store",
    "1.3.6.1.4.1.311.10.6.1": "Key Pack Licenses",
    "1.3.6.1.4.1.311.20.2.2": "Smart Card Logon",
    "1.3.6.1.5.2.3.5": "KDC Authentication",
    "1.3.6.1.5.5.7.3.7": "IP security use",
    "1.3.6.1.4.1.311.10.3.8": "Embedded Windows System Component Verification",
    "1.3.6.1.4.1.311.10.3.20": "Windows Kits Component",
    "1.3.6.1.5.5.7.3.6": "IP security tunnel termination",
    "1.3.6.1.4.1.311.10.3.5": "Windows Hardware Driver Verification",
    "1.3.6.1.5.5.8.2.2": "IP security IKE intermediate",
    "1.3.6.1.4.1.311.10.3.39": "Windows Hardware Driver Extended Verification",
    "1.3.6.1.4.1.311.10.6.2": "License Server Verification",
    "1.3.6.1.4.1.311.10.3.5.1": "Windows Hardware Driver Attested Verification",
    "1.3.6.1.4.1.311.76.5.1": "Dynamic Code Generato",
    "1.3.6.1.5.5.7.3.8": "Time Stamping",
    "1.3.6.1.4.1.311.10.3.4.1": "File Recovery",
    "1.3.6.1.4.1.311.2.6.1": "SpcRelaxedPEMarkerCheck",
    "2.23.133.8.1": "Endorsement Key Certificate",
    "1.3.6.1.4.1.311.2.6.2": "SpcEncryptedDigestRetryCount",
    "1.3.6.1.4.1.311.10.3.4": "Encrypting File System",
    "1.3.6.1.5.5.7.3.1": "Server Authentication",
    "1.3.6.1.4.1.311.61.5.1": "HAL Extension",
    "1.3.6.1.5.5.7.3.4": "Secure Email",
    "1.3.6.1.5.5.7.3.5": "IP security end system",
    "1.3.6.1.4.1.311.10.3.9": "Root List Signe",
    "1.3.6.1.4.1.311.10.3.30": "Disallowed List",
    "1.3.6.1.4.1.311.10.3.19": "Revoked List Signe",
    "1.3.6.1.4.1.311.10.3.21": "Windows RT Verification",
    "1.3.6.1.4.1.311.10.3.10": "Qualified Subordination",
    "1.3.6.1.4.1.311.10.3.12": "Document Signing",
    "1.3.6.1.4.1.311.10.3.24": "Protected Process Verification",
    "1.3.6.1.4.1.311.80.1": "Document Encryption",
    "1.3.6.1.4.1.311.10.3.22": "Protected Process Light Verification",
    "1.3.6.1.4.1.311.21.19": "Directory Service Email Replication",
    "1.3.6.1.4.1.311.21.5": "Private Key Archival",
    "1.3.6.1.4.1.311.10.5.1": "Digital Rights",
    "1.3.6.1.4.1.311.10.3.27": "Preview Build Signing",
    "1.3.6.1.4.1.311.20.2.1": "Certificate Request Agent",
    "2.23.133.8.2": "Platform Certificate",
    "1.3.6.1.4.1.311.20.1": "CTL Usage",
    "1.3.6.1.5.5.7.3.9": "OCSP Signing",
    "1.3.6.1.5.5.7.3.3": "Code Signing",
    "1.3.6.1.4.1.311.10.3.1": "Microsoft Trust List Signing",
    "1.3.6.1.4.1.311.10.3.2": "Microsoft Time Stamping",
    "1.3.6.1.4.1.311.76.8.1": "Microsoft Publishe",
    "1.3.6.1.5.5.7.3.2": "Client Authentication",
    "1.3.6.1.5.2.3.4": "PKINIT Client Authentication",
    "1.3.6.1.4.1.311.10.3.13": "Lifetime Signing",
    "2.5.29.37.0": "Any Purpose",
    "1.3.6.1.4.1.311.64.1.1": "Server Trust",
    "1.3.6.1.4.1.311.10.3.7": "OEM Windows System Component Verification",
}

__mspki_enrollment_flag = {
    0x00000001 : "INCLUDE_SYMMETRIC_ALGORITHMS",
    0x00000002 : "PEND_ALL_REQUESTS",
    0x00000004 : "PUBLISH_TO_KRA_CONTAINER",
    0x00000008 : "PUBLISH_TO_DS",
    0x00000010 : "AUTO_ENROLLMENT_CHECK_USER_DS_CERTIFICATE",
    0x00000020 : "AUTO_ENROLLMENT",
    0x00000080 : "CT_FLAG_DOMAIN_AUTHENTICATION_NOT_REQUIRED",
    0x00000040 : "PREVIOUS_APPROVAL_VALIDATE_REENROLLMENT",
    0x00000100 : "USER_INTERACTION_REQUIRED",
    0x00000200 : "ADD_TEMPLATE_NAME",
    0x00000400 : "REMOVE_INVALID_CERTIFICATE_FROM_PERSONAL_STORE",
    0x00000800 : "ALLOW_ENROLL_ON_BEHALF_OF",
    0x00001000 : "ADD_OCSP_NOCHECK",
    0x00002000 : "ENABLE_KEY_REUSE_ON_NT_TOKEN_KEYSET_STORAGE_FULL",
    0x00004000 : "NOREVOCATIONINFOINISSUEDCERTS",
    0x00008000 : "INCLUDE_BASIC_CONSTRAINTS_FOR_EE_CERTS",
    0x00010000 : "ALLOW_PREVIOUS_APPROVAL_KEYBASEDRENEWAL_VALIDATE_REENROLLMENT",
    0x00020000 : "ISSUANCE_POLICIES_FROM_REQUEST",
    0x00040000 : "SKIP_AUTO_RENEWAL",
    0x00080000 : "NO_SECURITY_EXTENSION"
}

def __format_flag(raw_value, flag_dict):
    try:
        int_value = int(raw_value)
    except ValueError:
        return raw_value

    parsed_flags = list()
    for flag, flag_label in flag_dict.items():
        if (int_value & flag) == flag:
            parsed_flags.append(flag_label)
    return parsed_flags

def __format_dict_lookup(raw_value, dictionary):
    try:
        return dictionary[int(raw_value)]
    except (ValueError, KeyError):
        return raw_value

def format_useraccountcontrol(raw_value):
    return __format_flag(raw_value, __uac_flags)

def format_samaccounttype(raw_value):
    return __sat_flags[int(raw_value)]

def format_ace_access_mask(raw_value):
    try:
        int_value = int(raw_value)
    except ValueError:
        return raw_value

    activedirectoryrights = list()
    for flag, flag_label in __access_mask_generic.items():
        if (int_value & flag) == flag:
            activedirectoryrights.append(flag_label)
            int_value ^= flag
    activedirectoryrights += __format_flag(raw_value, __access_mask)

    return activedirectoryrights


def format_managedpassword(raw_value):
    blob = MSDS_MANAGEDPASSWORD_BLOB()
    blob.fromString(raw_value)
    return binascii.hexlify(MD4.new(blob['CurrentPassword'][:-2]).digest()).decode('utf8')

def format_groupmsamembership(raw_value):
    sid = list()
    sr = SR_SECURITY_DESCRIPTOR(data=raw_value)
    for dacl in sr['Dacl']['Data']:
        sid.append(dacl['Ace']['Sid'].formatCanonical())
    return sid

def format_ace_flags(raw_value):
    return __format_flag(raw_value, __ace_flags)

def format_object_ace_flags(raw_value):
    return __format_flag(raw_value, __object_ace_flags)

def format_trustdirection(raw_value):
    return __format_dict_lookup(raw_value, __trust_direction)

def format_trusttype(raw_value):
    return __format_dict_lookup(raw_value, __trust_type)

def format_trustattributes(raw_value):
    return __format_flag(raw_value, __trust_attrib)

def format_ekus(raw_value):
    raw_value = raw_value.decode('utf-8')
    try:
        return __ekus[raw_value]
    except KeyError:
        return raw_value

def format_mspkienrollmentflag(raw_value):
    return __format_flag(raw_value, __mspki_enrollment_flag)



# This file is part of PywerView.

# PywerView is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PywerView is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PywerView.  If not, see <http://www.gnu.org/licenses/>.

# Yannick Méheut [yannick (at) meheut (dot) org] - Copyright © 2023

import inspect
import logging

from impacket.ldap.ldaptypes import ACE, ACCESS_ALLOWED_OBJECT_ACE, ACCESS_MASK, LDAP_SID, SR_SECURITY_DESCRIPTOR

import pywerview.functions.misc as misc

class ADObject:
    _well_known_sids = {'S-1-0-0': 'Nobody', 'S-1-0': 'Null Authority', 'S-1-1-0': 'Everyone',
                        'S-1-1': 'World Authority', 'S-1-2-0': 'Local', 'S-1-2-1': 'Console Logon',
                        'S-1-2': 'Local Authority', 'S-1-3-0': 'Creator Owner', 'S-1-3-1': 'Creator Group',
                        'S-1-3-2': 'Creator Owner Server', 'S-1-3-3': 'Creator Group Server', 'S-1-3-4': 'Owner Rights',
                        'S-1-3': 'Creator Authority', 'S-1-4': 'Non-unique Authority', 'S-1-5-10': 'Principal Self',
                        'S-1-5-11': 'Authenticated Users', 'S-1-5-12': 'Restricted Code', 'S-1-5-13': 'Terminal Server Users',
                        'S-1-5-14': 'Remote Interactive Logon', 'S-1-5-17': 'This Organization', 'S-1-5-18': 'Local System',
                        'S-1-5-19': 'NT Authority', 'S-1-5-1': 'Dialup', 'S-1-5-20': 'NT Authority', 'S-1-5-2': 'Network',
                        'S-1-5-32-544': 'Administrators' ,'S-1-5-32-546': 'Guests', 'S-1-5-32-547': 'Power Users',
                        'S-1-5-32-548': 'Account Operators', 'S-1-5-32-549': 'Server Operators',
                        'S-1-5-32-550': 'Print Operators', 'S-1-5-32-551': 'Backup Operators', 'S-1-5-32-552': 'Replicators',
                        'S-1-5-32-554': 'Builtin\\Pre-Windows 2000 Compatible Access',
                        'S-1-5-32-555': 'Builtin\\Remote Desktop Users',
                        'S-1-5-32-556': 'Builtin\\Network Configuration Operators',
                        'S-1-5-32-557': 'Builtin\\Incoming Forest Trust Builders',
                        'S-1-5-32-558': 'Builtin\\Performance Monitor Users',
                        'S-1-5-32-559': 'Builtin\\Performance Log Users',
                        'S-1-5-32-560': 'Builtin\\Windows Authorization Access Group',
                        'S-1-5-32-561': 'Builtin\\Terminal Server License Servers',
                        'S-1-5-32-562': 'Builtin\\Distributed COM Users',
                        'S-1-5-32-568': 'Builtin\\IIS_IUSRS',
                        'S-1-5-32-569': 'Builtin\\Cryptographic Operators',
                        'S-1-5-32-573': 'Builtin\\Event Log Readers',
                        'S-1-5-32-574': 'Builtin\\Certificate Service DCOM Access',
                        'S-1-5-32-575': 'Builtin\\RDS Remote Access Servers',
                        'S-1-5-32-576': 'Builtin\\RDS Endpoint Servers',
                        'S-1-5-32-577': 'Builtin\\RDS Management Servers',
                        'S-1-5-32-578': 'Builtin\\Hyper-V Administrators',
                        'S-1-5-32-579': 'Builtin\\Access Control Assistance Operators',
                        'S-1-5-32-580': 'Builtin\\Remote Management Users',
                        'S-1-5-32-582': 'Storage Replica Administrators',
                        'S-1-5-3': 'Batch', 'S-1-5-4': 'Interactive', 'S-1-5-64-10': 'NTLM Authentication',
                        'S-1-5-64-14': 'SChannel Authentication', 'S-1-5-64-21': 'Digest Authentication',
                        'S-1-5-6': 'Service', 'S-1-5-7': 'Anonymous', 'S-1-5-80-0': 'NT Services\\All Services',
                        'S-1-5-80': 'NT Service', 'S-1-5-8': 'Proxy', 'S-1-5-9': 'Enterprise Domain Controllers',
                        'S-1-5': 'NT Authority'}

    _well_known_rids = { '498' : 'Enterprise Read-only Domain Controllers',
                         '512' : 'Domain Admins', '513' : 'Domain Users', '514' : 'Domain Guests', '515' : 'Domain Computers',
                         '516' : 'Domain Controllers', '517' : 'Cert Publishers', '519' : 'Enterprise Admins',
                         '520' : 'Group Policy Creator Owners', '522' : 'Cloneable Domain COntrollers', '526' : 'Key Admins',
                         '527' : 'Enterptise Key Admins', '553' : 'RAS and IAS Servers'}

    def __init__(self, attributes):
        logger = logging.getLogger('pywerview_main_logger.ADObject')
        logger.ULTRA = 5
        self._logger = logger

        self._attributes_dict = dict()
        self.add_attributes(attributes)

    def add_attributes(self, attributes):
        self._logger.log(self._logger.ULTRA,'ADObject instancied with the following attributes : {}'.format(attributes))
        for attr in attributes:
            self._attributes_dict[attr.lower()] = attributes[attr]

    def __getattr__(self, attr):
        try:
            return self._attributes_dict[attr]
        except KeyError:
            if attr == 'isgroup':
                try:
                    return 'group' in self._attributes_dict['objectclass']
                except KeyError:
                    return False
            raise AttributeError

    # In this method, we try to pretty print common AD attributes
    def __str__(self):
        s = str()
        max_length = 0
        for attr in self._attributes_dict:
            if len(attr) > max_length:
                max_length = len(attr)
        for attr in self._attributes_dict:
            attribute = self._attributes_dict[attr]
            self._logger.log(self._logger.ULTRA,'Trying to print : attribute name = {0} / value = {1}'.format(attr, attribute))
            if isinstance(attribute, list):
                if any(isinstance(x, bytes) for x in attribute):
                    attribute = ['{}...'.format(x.hex()[:97]) for x in attribute]
                attribute_temp = ', '.join(str(x) for x in attribute)
                if len(attribute_temp) > 100:
                    attribute_temp = str()
                    line = str()
                    for x in attribute:
                        if len(line) + len(str(x)) <= 100:
                            line += '{}, '.format(x)
                        else:
                            attribute_temp += line + '\n' + ' ' * (max_length + 2)
                            line = str()
                            line += '{}, '.format(x)
                    attribute_temp += line + '\n' + ' ' * (max_length + 2)
                attribute = attribute_temp.rstrip().rstrip(',')
            elif isinstance(attribute, bytes):
                attribute = '{}...'.format(attribute.hex()[:100])
            elif isinstance(attribute, ADObject):
                attribute = ('\n' + str(attribute)).replace('\n', '\n\t')

            s += '{}: {}{}\n'.format(attr, ' ' * (max_length - len(attr)), attribute)

        s = s[:-1]
        return s

    def __repr__(self):
        return str(self)

    def to_json(self):
        return self._attributes_dict

class ACE(ADObject):

    def __init__(self, attributes):
        ADObject.__init__(self, attributes)

        # We set iscallback, depending on the type of ACE
        self._attributes_dict['iscallbak'] = ('CALLBACK' in self.acetype)

class User(ADObject):
    pass

class Group(ADObject):
    pass

class Computer(ADObject):
    pass

class FileServer(ADObject):
    pass

class DFS(ADObject):
    pass

class OU(ADObject):
    pass

class Site(ADObject):
    pass

class Subnet(ADObject):
    pass

class Trust(ADObject):
    pass

class GPO(ADObject):
    pass

class PSO(ADObject):
    pass

class GptTmpl(ADObject):
    def to_json(self):
        json_dict = {}
        for k, v in self._attributes_dict.items():
            json_dict[k] = v.to_json()
        return json_dict

class GPOGroup(ADObject):
    pass

class Policy(ADObject):
    pass

class GPOComputerAdmin(ADObject):
    pass

class GPOLocation(ADObject):
    pass

class GMSAAccount(ADObject):
    pass

class SMSAAccount(ADObject):
    pass

class ObjectOwner(ADObject):
    pass

class PKIEnrollmentService(ADObject):
    _enrollment_uuids = { "All-Extended-Rights" : "00000000-0000-0000-0000-000000000000",
                          "Certificate-Enrollment" : "0e10c968-78fb-11d2-90d4-00c04f79dc55",
                          "Certificate-AutoEnrollment" : "a05b8cc2-17bc-4802-a710-e7c15ab866a2"}
    pass

class PKICertificateTemplate(PKIEnrollmentService):
    pass



#!/usr/bin/env python3
#
# This file is part of PywerView.

# PywerView is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PywerView is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PywerView.  If not, see <http://www.gnu.org/licenses/>.

# Yannick Méheut [yannick (at) meheut (dot) org] - Copyright © 2023

import logging
import argparse
import json
import datetime
from pywerview.cli.helpers import *
from pywerview.functions.hunting import *

def main():
    # Main parser
    parser = argparse.ArgumentParser(description='Rewriting of some PowerView\'s functionalities in Python')
    subparsers = parser.add_subparsers(title='Subcommands', description='Available subcommands', dest='submodule')

    # hack for python < 3.9 : https://stackoverflow.com/questions/23349349/argparse-with-required-subparser
    subparsers.required = True

    # Logging parser
    logging_parser = argparse.ArgumentParser(add_help=False)
    logging_parser.add_argument('-l', '--logging-level', dest='logging_level', type=str.upper,
            choices=['CRITICAL', 'WARNING', 'DEBUG', 'ULTRA'], default='CRITICAL',
            help='SDTERR logging level: '
                 'CRITICAL: Only critical errors are displayed (default), '
                 'WARNING: Warnings are displayed, along with citical errors, '
                 'DEBUG: Debug level (caution: very verbose), '
                 'ULTRA: Extreme debugging level (caution: very very verbose)')
    
    # json parser
    json_output_parser = argparse.ArgumentParser(add_help=False)
    json_output_parser.add_argument('--json', dest='json_output', action='store_true',
            help='Print results in JSON format')

    # Credentials parser
    credentials_parser = argparse.ArgumentParser(add_help=False)
    credentials_parser.add_argument('-w', '--workgroup', dest='domain',
            default=str(), help='Name of the domain we authenticate with')
    credentials_parser.add_argument('-u', '--user',
            help='Username used to connect to the Domain Controller')
    credentials_parser.add_argument('-p', '--password',
            help='Password associated to the username')
    credentials_parser.add_argument('--hashes', action='store', metavar='LMHASH:NTHASH',
            help='NTLM hashes, format is [LMHASH:]NTHASH')
    credentials_parser.add_argument('-k', action='store_true', dest='do_kerberos',
            help='Use Kerberos authentication. Grabs credentials from ccache file '
            '(KRB5CCNAME) based on target parameters. If valid credentials '
            'cannot be found, it will use the ones specified in the command '
            'line')

    # Certificate authentication parser
    certificate_parser = argparse.ArgumentParser(add_help=False)
    certificate_parser.add_argument('--cert', dest='user_cert',
            help='Certificatie associated to the user')
    certificate_parser.add_argument('--key', dest='user_key',
            help='Private key associated to the user')

    # SIMPLE authentication parser
    simple_auth_parser = argparse.ArgumentParser(add_help=False)
    simple_auth_parser.add_argument('--simple-auth', action='store_true', dest='do_simple',
             help='Force SIMPLE LDAP authentication')


    # AD parser, used for net* functions running against a domain controller
    ad_parser = argparse.ArgumentParser(add_help=False, parents=[credentials_parser])
    ad_parser.add_argument('-t', '--dc-ip', dest='domain_controller',
            required=True, help='IP address of the Domain Controller to target')
    ad_parser.add_argument('--tls', action='store_true', dest='do_tls',
            help='Force TLS connection to the Domain Controller')

    # Target parser, used for net* functions running against a normal computer
    target_parser = argparse.ArgumentParser(add_help=False, parents=[credentials_parser])
    target_parser.add_argument('--computername', dest='target_computername',
            required=True, help='IP address of the computer target')

    # Hunter parser, used for hunting functions
    hunter_parser = argparse.ArgumentParser(add_help=False)
    hunter_parser.add_argument('--computername', dest='queried_computername',
            nargs='+', default=list(), help='Host to enumerate against')
    hunter_parser.add_argument('--computerfile', dest='queried_computerfile',
            type=argparse.FileType('r'), help='File of hostnames/IPs to search')
    hunter_parser.add_argument('--computer-filter', dest='queried_computerfilter',
            type=str, default=str(), help='Custom filter used to search computers against the DC')
    hunter_parser.add_argument('--computer-adspath', dest='queried_computeradspath',
            type=str, default=str(), help='ADS path used to search computers against the DC')
    hunter_parser.add_argument('--groupname', dest='queried_groupname',
            help='Group name to query for target users')
    hunter_parser.add_argument('--targetserver', dest='target_server',
            help='Hunt for users who are effective local admins on this target server')
    hunter_parser.add_argument('--username', dest='queried_username',
            help='Hunt for a specific user name')
    hunter_parser.add_argument('--user-filter', dest='queried_userfilter',
            type=str, default=str(), help='Custom filter used to search users against the DC')
    hunter_parser.add_argument('--user-adspath', dest='queried_useradspath',
            type=str, default=str(), help='ADS path used to search users against the DC')
    hunter_parser.add_argument('--userfile', dest='queried_userfile',
            type=argparse.FileType('r'), help='File of user names to target')
    hunter_parser.add_argument('--threads', type=int,
            default=1, help='Number of threads to use (default: %(default)s)')
    hunter_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query for machines')

    # Parser for the get-adobject command
    get_adobject_parser = subparsers.add_parser('get-adobject', help='Takes a domain SID, '\
        'samAccountName or name, and return the associated object',
        parents=[ad_parser, logging_parser, json_output_parser, certificate_parser, simple_auth_parser])
    get_adobject_parser.add_argument('--sid', dest='queried_sid',
            help='SID to query (wildcards accepted)')
    get_adobject_parser.add_argument('--sam-account-name', dest='queried_sam_account_name',
            help='samAccountName to query (wildcards accepted)')
    get_adobject_parser.add_argument('--name', dest='queried_name',
            help='Name to query (wildcards accepted)')
    get_adobject_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    get_adobject_parser.add_argument('-a', '--ads-path',
            help='Additional ADS path')
    get_adobject_parser.add_argument('--custom-filter', dest='custom_filter',
            default=str(), help='Custom filter')
    get_adobject_parser.add_argument('--attributes', nargs='+', dest='attributes',
            default=[], help='Object attributes to return')
    get_adobject_parser.set_defaults(func=get_adobject)

    # Parser for the get-objectowner command
    get_objectowner_parser = subparsers.add_parser('get-objectowner', help='Takes a domain SID, '\
        'samAccountName or name, and return the associated object owner',
        parents=[ad_parser, logging_parser, json_output_parser, certificate_parser, simple_auth_parser])
    get_objectowner_parser.add_argument('--sid', dest='queried_sid',
            help='SID to query (wildcards accepted)')
    get_objectowner_parser.add_argument('--sam-account-name', dest='queried_sam_account_name',
            help='samAccountName to query (wildcards accepted)')
    get_objectowner_parser.add_argument('--name', dest='queried_name',
            help='Name to query (wildcards accepted)')
    get_objectowner_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    get_objectowner_parser.add_argument('-a', '--ads-path',
            help='Additional ADS path')
    get_objectowner_parser.add_argument('--custom-filter', dest='custom_filter',
            default=str(), help='Custom filter')
    get_objectowner_parser.add_argument('--resolve-sids', dest='resolve_sids',
            action='store_true', help='Resolve SIDs')
    get_objectowner_parser.set_defaults(func=get_objectowner)

    # Parser for the get-netpki command
    get_netpki_parser = subparsers.add_parser('get-netpki', help='Returns a list of all the '\
         'pKIEnrollmentService objects',
        parents=[ad_parser, logging_parser, json_output_parser, certificate_parser, simple_auth_parser])
    get_netpki_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    get_netpki_parser.add_argument('--resolve-sids', dest='resolve_sids',
            action='store_true', help='Resolve SIDs')
    get_netpki_parser.add_argument('--full-data', action='store_true',
            help='If set, returns full information on the pki')
    get_netpki_parser.add_argument('--caname', dest='queried_ca_name',
        help='CA name to query (wildcards accepted)')
    get_netpki_parser.set_defaults(func=get_netpki)

    # Parser for the get-netcerttmpl command
    get_netcerttmpl_parser = subparsers.add_parser('get-netcerttmpl', help='Returns a list of all the '\
         'pKICertificateTemplate objects',
        parents=[ad_parser, logging_parser, json_output_parser, certificate_parser, simple_auth_parser])
    get_netcerttmpl_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    get_netcerttmpl_parser.add_argument('--resolve-sids', dest='resolve_sids',
            action='store_true', help='Resolve SIDs')
    get_netcerttmpl_parser.add_argument('--caname', dest='queried_ca_name',
        help='CA name to query')
    get_netcerttmpl_parser.set_defaults(func=get_netcerttmpl)

    # Parser for the get-netgmsa command
    get_netgmsa_parser = subparsers.add_parser('get-netgmsa', help='Returns a list of all the '\
        'gMSA of the specified domain. To retrieve passwords, you need a privileged account and '\
        'a TLS connection to the LDAP server (use the --tls switch).',
        parents=[ad_parser, logging_parser, json_output_parser, certificate_parser, simple_auth_parser])
    get_netgmsa_parser.add_argument('--sid', dest='queried_sid',
            help='SID to query (wildcards accepted)')
    get_netgmsa_parser.add_argument('--sam-account-name', dest='queried_sam_account_name',
            help='samAccountName to query (wildcards accepted)')
    get_netgmsa_parser.add_argument('--name', dest='queried_name',
            help='Name to query (wildcards accepted)')
    get_netgmsa_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    get_netgmsa_parser.add_argument('-a', '--ads-path',
            help='Additional ADS path')
    get_netgmsa_parser.add_argument('--resolve-sids', dest='resolve_sids',
            action='store_true', help='Resolve SIDs when querying PrincipalsAllowedToRetrieveManagedPassword')
    get_netgmsa_parser.set_defaults(func=get_netgmsa)

    # Parser for the get-netsmsa command
    get_netsmsa_parser = subparsers.add_parser('get-netsmsa', help='Returns a list of all the '\
        'sMSA of the specified domain.',
        parents=[ad_parser, logging_parser, json_output_parser, certificate_parser, simple_auth_parser])
    get_netsmsa_parser.add_argument('--sid', dest='queried_sid',
            help='SID to query (wildcards accepted)')
    get_netsmsa_parser.add_argument('--sam-account-name', dest='queried_sam_account_name',
            help='samAccountName to query (wildcards accepted)')
    get_netsmsa_parser.add_argument('--name', dest='queried_name',
            help='Name to query (wildcards accepted)')
    get_netsmsa_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    get_netsmsa_parser.add_argument('-a', '--ads-path',
            help='Additional ADS path')
    get_netsmsa_parser.set_defaults(func=get_netsmsa)
    
    # Parser for the get-objectacl command
    get_objectacl_parser = subparsers.add_parser('get-objectacl', help='Takes a domain SID, '\
        'samAccountName or name, and return the ACL of the associated object',
        parents=[ad_parser, logging_parser, json_output_parser, certificate_parser, simple_auth_parser])
    get_objectacl_parser.add_argument('--sid', dest='queried_sid',
            help='SID to query (wildcards accepted)')
    get_objectacl_parser.add_argument('--sam-account-name', dest='queried_sam_account_name',
            help='samAccountName to query (wildcards accepted)')
    get_objectacl_parser.add_argument('--name', dest='queried_name',
            help='Name to query (wildcards accepted)')
    get_objectacl_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    get_objectacl_parser.add_argument('-a', '--ads-path',
            help='Additional ADS path')
    get_objectacl_parser.add_argument('--sacl', action='store_true',
            help='Return the SACL instead of the DACL for the object (requires '\
            'a privileged account)')
    get_objectacl_parser.add_argument('--rights-filter', dest='rights_filter',
            choices=['reset-password', 'write-members', 'allowed-to-authenticate', 'all'], help='A specific set of rights to return '\
                    '(reset-password, write-members, allowed-to-authenticate , all)')
    get_objectacl_parser.add_argument('--resolve-sids', dest='resolve_sids',
            action='store_true', help='Resolve SIDs when querying an ACL')
    get_objectacl_parser.add_argument('--resolve-guids', action='store_true',
            help='Resolve GUIDs to their display names')
    get_objectacl_parser.add_argument('--custom-filter', dest='custom_filter',
            default=str(), help='Custom filter')
    get_objectacl_parser.set_defaults(func=get_objectacl)

    # Parser for the get-netuser command
    get_netuser_parser = subparsers.add_parser('get-netuser', help='Queries information about '\
        'a domain user', parents=[ad_parser, logging_parser, json_output_parser,
                                  certificate_parser, simple_auth_parser])
    get_netuser_parser.add_argument('--username', dest='queried_username',
            help='Username to query (wildcards accepted)')
    get_netuser_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    get_netuser_parser.add_argument('-a', '--ads-path',
            help='Additional ADS path')
    get_netuser_parser.add_argument('--unconstrained', action='store_true',
            help='Query only users with unconstrained delegation')
    get_netuser_parser.add_argument('--admin-count', action='store_true',
            help='Query only users with adminCount=1')
    get_netuser_parser.add_argument('--allow-delegation', action='store_true',
            help='Return user accounts that are not marked as \'sensitive and not allowed for delegation\'')
    get_netuser_parser.add_argument('--preauth-notreq', action='store_true',
            help='Search for users with the PREAUTH_NOT_REQUIRED account control')
    get_netuser_parser.add_argument('--spn', action='store_true',
            help='Query only users with not-null Service Principal Names')
    get_netuser_parser.add_argument('--custom-filter', dest='custom_filter',
            default=str(), help='Custom filter')
    get_netuser_parser.add_argument('--attributes', nargs='+', dest='attributes',
            default=[], help='Object attributes to return')
    get_netuser_parser.set_defaults(func=get_netuser)

    # Parser for the get-netgroup command
    get_netgroup_parser = subparsers.add_parser('get-netgroup', help='Get a list of all current '\
        'domain groups, or a list of groups a domain user is member of',
        parents=[ad_parser, logging_parser, json_output_parser, certificate_parser, simple_auth_parser])
    get_netgroup_parser.add_argument('--groupname', dest='queried_groupname',
            default='*', help='Group to query (wildcards accepted)')
    get_netgroup_parser.add_argument('--sid', dest='queried_sid',
            help='Group SID to query')
    get_netgroup_parser.add_argument('--username', dest='queried_username',
            help='Username to query: will list the groups this user is a member of (wildcards accepted)')
    get_netgroup_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    get_netgroup_parser.add_argument('-a', '--ads-path', dest='ads_path',
            help='Additional ADS path')
    get_netgroup_parser.add_argument('--full-data', action='store_true',
            help='If set, returns full information on the groups, otherwise, just the samAccountName')
    get_netgroup_parser.add_argument('--admin-count', action='store_true',
            help='Query only users with adminCount=1')
    get_netgroup_parser.add_argument('--custom-filter', dest='custom_filter',
            default=str(), help='Custom filter')
    get_netgroup_parser.set_defaults(func=get_netgroup)

    # Parser for the get-netcomputer command
    get_netcomputer_parser = subparsers.add_parser('get-netcomputer', help='Queries informations about '\
        'domain computers', parents=[ad_parser, logging_parser, json_output_parser,
                                     certificate_parser, simple_auth_parser])
    get_netcomputer_parser.add_argument('--computername', dest='queried_computername',
            default=str(), help='Computer name to query')
    get_netcomputer_parser.add_argument('-os', '--operating-system', dest='queried_os',
            help='Return computers with a specific operating system (wildcards accepted)')
    get_netcomputer_parser.add_argument('-sp', '--service-pack', dest='queried_sp',
            help='Return computers with a specific service pack (wildcards accepted)')
    get_netcomputer_parser.add_argument('-spn', '--service-principal-name', dest='queried_spn',
            help='Return computers with a specific service principal name (wildcards accepted)')
    get_netcomputer_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    get_netcomputer_parser.add_argument('-a', '--ads-path', dest='ads_path',
            help='Additional ADS path')
    get_netcomputer_parser.add_argument('--printers', action='store_true',
            help='Query only printers')
    get_netcomputer_parser.add_argument('--unconstrained', action='store_true',
            help='Query only computers with unconstrained delegation')
    get_netcomputer_parser.add_argument('--laps-passwords', action='store_true', dest='laps_passwords',
            help='Query only computers for which the user can read LAPS passwords')
    # See: https://www.trustedsec.com/blog/diving-into-pre-created-computer-accounts/
    get_netcomputer_parser.add_argument('--pre-created', action='store_true', dest='pre_created',
            help='Query only computers which are potentially vulnerable to "pre-created computer account" attack. '\
                 'Caution: This option is prone to false positives and negatives.')
    get_netcomputer_parser.add_argument('--ping', action='store_true',
            help='Ping computers (will only return up computers)')
    get_netcomputer_parser.add_argument('--full-data', action='store_true',
            help='If set, returns full information on the groups, otherwise, just the dnsHostName')
    get_netcomputer_parser.add_argument('--custom-filter', dest='custom_filter',
            default=str(), help='Custom filter')
    get_netcomputer_parser.add_argument('--attributes', nargs='+', dest='attributes',
            default=[], help='Object attributes to return')
    get_netcomputer_parser.set_defaults(func=get_netcomputer)

    # Parser for the get-netdomaincontroller command
    get_netdomaincontroller_parser = subparsers.add_parser('get-netdomaincontroller', help='Get a list of '\
        'domain controllers for the given domain', parents=[ad_parser, logging_parser, 
                                                            json_output_parser, certificate_parser,
                                                            simple_auth_parser])
    get_netdomaincontroller_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    get_netdomaincontroller_parser.set_defaults(func=get_netdomaincontroller)

    # Parser for the get-netfileserver command
    get_netfileserver_parser = subparsers.add_parser('get-netfileserver', help='Return a list of '\
        'file servers, extracted from the domain users\' homeDirectory, scriptPath, and profilePath fields',
        parents=[ad_parser, logging_parser, json_output_parser, certificate_parser, simple_auth_parser])
    get_netfileserver_parser.add_argument('--target-users', nargs='+',
            metavar='TARGET_USER', help='A list of users to target to find file servers (wildcards accepted)')
    get_netfileserver_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    get_netfileserver_parser.set_defaults(func=get_netfileserver)

    # Parser for the get-dfsshare command
    get_dfsshare_parser = subparsers.add_parser('get-dfsshare', help='Return a list of '\
        'all fault tolerant distributed file systems for a given domain', parents=[ad_parser, logging_parser, json_output_parser])
    get_dfsshare_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    get_dfsshare_parser.add_argument('-v', '--version', nargs='+', choices=['v1', 'v2'],
            default=['v1', 'v2'], help='The version of DFS to query for servers: v1, v2 or all (default: all)')
    get_dfsshare_parser.add_argument('-a', '--ads-path', dest='ads_path',
            help='Additional ADS path')
    get_dfsshare_parser.set_defaults(func=get_dfsshare)

    # Parser for the get-netou command
    get_netou_parser = subparsers.add_parser('get-netou', help='Get a list of all current '\
        'OUs in the domain', parents=[ad_parser, logging_parser, json_output_parser,
                                      certificate_parser, simple_auth_parser])
    get_netou_parser.add_argument('--ouname', dest='queried_ouname',
            default='*', help='OU name to query (wildcards accepted)')
    get_netou_parser.add_argument('--guid', dest='queried_guid',
            help='Only return OUs with the specified GUID in their gplink property.')
    get_netou_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    get_netou_parser.add_argument('-a', '--ads-path',
            help='Additional ADS path')
    get_netou_parser.add_argument('--full-data', action='store_true',
            help='If set, returns full information on the OUs, otherwise, just the adspath')
    get_netou_parser.set_defaults(func=get_netou)

    # Parser for the get-netsite command
    get_netsite_parser = subparsers.add_parser('get-netsite', help='Get a list of all current '\
        'sites in the domain', parents=[ad_parser, logging_parser, json_output_parser,
                                        certificate_parser, simple_auth_parser])
    get_netsite_parser.add_argument('--sitename', dest='queried_sitename',
            help='Site name to query (wildcards accepted)')
    get_netsite_parser.add_argument('--guid', dest='queried_guid',
            help='Only return sites with the specified GUID in their gplink property.')
    get_netsite_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    get_netsite_parser.add_argument('-a', '--ads-path',
            help='Additional ADS path')
    get_netsite_parser.add_argument('--full-data', action='store_true',
            help='If set, returns full information on the sites, otherwise, just the name')
    get_netsite_parser.set_defaults(func=get_netsite)

    # Parser for the get-netsubnet command
    get_netsubnet_parser = subparsers.add_parser('get-netsubnet', help='Get a list of all current '\
        'subnets in the domain', parents=[ad_parser, logging_parser, json_output_parser,
                                          certificate_parser, simple_auth_parser])
    get_netsubnet_parser.add_argument('--sitename', dest='queried_sitename',
            help='Only return subnets for the specified site name (wildcards accepted)')
    get_netsubnet_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    get_netsubnet_parser.add_argument('-a', '--ads-path',
            help='Additional ADS path')
    get_netsubnet_parser.add_argument('--full-data', action='store_true',
            help='If set, returns full information on the subnets, otherwise, just the name')
    get_netsubnet_parser.set_defaults(func=get_netsubnet)

    # Parser for the get-netdomaintrust command
    get_netdomaintrust_parser = subparsers.add_parser('get-netdomaintrust', help='Returns a list of all the '\
        'trusts of the specified domain', parents=[ad_parser, logging_parser, json_output_parser,
                                                   certificate_parser, simple_auth_parser])
    get_netdomaintrust_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    get_netdomaintrust_parser.add_argument('--full-data', action='store_true',
            help='If set, returns full information on the trusts, otherwise, just basic info')
    get_netdomaintrust_parser.set_defaults(func=get_netdomaintrust)

    # Parser for the get-netgpo command
    get_netgpo_parser = subparsers.add_parser('get-netgpo', help='Get a list of all current '\
        'GPOs in the domain', parents=[ad_parser, logging_parser, json_output_parser,
                                       certificate_parser, simple_auth_parser])
    get_netgpo_parser.add_argument('--gponame', dest='queried_gponame',
            default='*', help='GPO name to query for (wildcards accepted)')
    get_netgpo_parser.add_argument('--displayname', dest='queried_displayname',
            help='Display name to query for (wildcards accepted)')
    get_netgpo_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    get_netgpo_parser.add_argument('-a', '--ads-path',
            help='Additional ADS path')
    get_netgpo_parser.set_defaults(func=get_netgpo)

    # Parser for the get-netpso command
    get_netpso_parser = subparsers.add_parser('get-netpso', help='Get a list of all current '\
        'PSOs in the domain', parents=[ad_parser, logging_parser, json_output_parser,
                                       certificate_parser, simple_auth_parser])
    get_netpso_parser.add_argument('--psoname', dest='queried_psoname',
            default='*', help='pso name to query for (wildcards accepted)')
    get_netpso_parser.add_argument('--displayname', dest='queried_displayname',
            help='Display name to query for (wildcards accepted)')
    get_netpso_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    get_netpso_parser.add_argument('-a', '--ads-path',
            help='Additional ADS path')
    get_netpso_parser.set_defaults(func=get_netpso)

    # Parser for the get-domainpolicy command
    get_domainpolicy_parser = subparsers.add_parser('get-domainpolicy', help='Returns the default domain or DC '\
        'policy for the queried domain or DC', parents=[ad_parser, logging_parser, json_output_parser])
    get_domainpolicy_parser.add_argument('--source', dest='source', default='domain',
            choices=['domain', 'dc'], help='Extract domain or DC policy (default: %(default)s)')
    get_domainpolicy_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    get_domainpolicy_parser.add_argument('--resolve-sids', dest='resolve_sids',
            action='store_true', help='Resolve SIDs when querying a DC policy')
    get_domainpolicy_parser.set_defaults(func=get_domainpolicy)

    # Parser for the get-gpttmpl command
    get_gpttmpl_parser = subparsers.add_parser('get-gpttmpl', help='Helper to parse a GptTmpl.inf policy '\
            'file path into a custom object', parents=[ad_parser, logging_parser, json_output_parser])
    get_gpttmpl_parser.add_argument('--gpt-tmpl-path', type=str, required=True,
            dest='gpttmpl_path', help='The GptTmpl.inf file path name to parse')
    get_gpttmpl_parser.set_defaults(func=get_gpttmpl)

    # Parser for the get-netgpogroup command
    get_netgpogroup_parser = subparsers.add_parser('get-netgpogroup', help='Parses all GPOs in the domain '\
        'that set "Restricted Group" or "Groups.xml"', parents=[ad_parser, logging_parser, json_output_parser])
    get_netgpogroup_parser.add_argument('--gponame', dest='queried_gponame',
            default='*', help='GPO name to query for (wildcards accepted)')
    get_netgpogroup_parser.add_argument('--displayname', dest='queried_displayname',
            help='Display name to query for (wildcards accepted)')
    get_netgpogroup_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    get_netgpogroup_parser.add_argument('-a', '--ads-path',
            help='Additional ADS path')
    get_netgpogroup_parser.add_argument('--resolve-sids', dest='resolve_sids',
            action='store_true', help='Resolve SIDs of the members and the target groups')
    get_netgpogroup_parser.set_defaults(func=get_netgpogroup)

    # Parser for the find-gpocomputeradmin command
    find_gpocomputeradmin_parser = subparsers.add_parser('find-gpocomputeradmin', help='Takes a computer (or OU) and determine '\
        'who has administrative access to it via GPO', parents=[ad_parser, logging_parser, json_output_parser])
    find_gpocomputeradmin_parser.add_argument('--computername', dest='queried_computername',
            default=str(), help='The computer to determine who has administrative access to it')
    find_gpocomputeradmin_parser.add_argument('--ouname', dest='queried_ouname',
            default=str(), help='OU name to determine who has administrative access to computers within it')
    find_gpocomputeradmin_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    find_gpocomputeradmin_parser.add_argument('-r', '--recurse', dest='recurse',
            action='store_true', help='If one of the returned members is a group, '\
                    'recurse and get all members')
    find_gpocomputeradmin_parser.set_defaults(func=find_gpocomputeradmin)

    # Parser for the find-gpolocation command
    find_gpolocation_parser = subparsers.add_parser('find-gpolocation', help='Takes a username or a group name and determine '\
        'the computers it has administrative access to via GPO', parents=[ad_parser, logging_parser, json_output_parser])
    find_gpolocation_parser.add_argument('--username', dest='queried_username',
            default=str(), help='The username to query for access (no wildcard)')
    find_gpolocation_parser.add_argument('--groupname', dest='queried_groupname',
            default=str(), help='The group name to query for access (no wildcard)')
    find_gpolocation_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    find_gpolocation_parser.add_argument('--local-group', dest='queried_localgroup',
            default='S-1-5-32-544', help='The local group to check access against. It can be ' \
                    '\'Administrators\', \'RDP\', or a \'S-1-5-X\' SID type')
    find_gpolocation_parser.set_defaults(func=find_gpolocation)

    # Parser for the get-netgroupmember command
    get_netgroupmember_parser = subparsers.add_parser('get-netgroupmember', help='Return a list of members of a domain group',
        parents=[ad_parser, logging_parser, json_output_parser, certificate_parser, simple_auth_parser])
    get_netgroupmember_parser.add_argument('--groupname', dest='queried_groupname',
            help='Group to query, defaults to the \'Domain Admins\' group (wildcards accepted)')
    get_netgroupmember_parser.add_argument('--sid', dest='queried_sid',
            help='SID to query')
    get_netgroupmember_parser.add_argument('-d', '--domain', dest='queried_domain',
            help='Domain to query')
    get_netgroupmember_parser.add_argument('-a', '--ads-path', dest='ads_path',
            help='Additional ADS path')
    get_netgroupmember_parser.add_argument('-r', '--recurse', action='store_true',
            help='If the group member is a group, try to resolve its members as well')
    get_netgroupmember_parser.add_argument('--use-matching-rule', action='store_true',
            help='Use LDAP_MATCHING_RULE_IN_CHAIN in the LDAP search query when -Recurse is specified.\n' \
        'Much faster than manual recursion, but doesn\'t reveal cross-domain groups')
    get_netgroupmember_parser.add_argument('--full-data', action='store_true',
            help='If set, returns full information on the members')
    get_netgroupmember_parser.add_argument('--custom-filter', dest='custom_filter',
            default=str(), help='Custom filter')
    get_netgroupmember_parser.set_defaults(func=get_netgroupmember)

    # Parser for the get-netsession command
    get_netsession_parser = subparsers.add_parser('get-netsession', help='Queries a host to return a '\
        'list of active sessions on the host (you can use local credentials instead of domain credentials)',
        parents=[target_parser, logging_parser, json_output_parser])
    get_netsession_parser.set_defaults(func=get_netsession)

    #Parser for the get-localdisks command
    get_localdisks_parser = subparsers.add_parser('get-localdisks', help='Queries a host to return a '\
        'list of active disks on the host (you can use local credentials instead of domain credentials)',
        parents=[target_parser, logging_parser, json_output_parser])
    get_localdisks_parser.set_defaults(func=get_localdisks)

    #Parser for the get-netdomain command
    get_netdomain_parser = subparsers.add_parser('get-netdomain', help='Queries a host for available domains',
        parents=[ad_parser, logging_parser, json_output_parser])
    get_netdomain_parser.set_defaults(func=get_netdomain)

    # Parser for the get-netshare command
    get_netshare_parser = subparsers.add_parser('get-netshare', help='Queries a host to return a '\
        'list of available shares on the host (you can use local credentials instead of domain credentials)',
        parents=[target_parser, logging_parser, json_output_parser])
    get_netshare_parser.set_defaults(func=get_netshare)

    # Parser for the get-netloggedon command
    get_netloggedon_parser = subparsers.add_parser('get-netloggedon', help='This function will '\
        'execute the NetWkstaUserEnum RPC call to query a given host for actively logged on '\
        'users', parents=[target_parser, logging_parser, json_output_parser])
    get_netloggedon_parser.set_defaults(func=get_netloggedon)

    # Parser for the get-netlocalgroup command
    get_netlocalgroup_parser = subparsers.add_parser('get-netlocalgroup', help='Gets a list of '\
        'members of a local group on a machine, or returns every local group. You can use local '\
        'credentials instead of domain credentials, however, domain credentials are needed to '\
        'resolve domain SIDs.', parents=[target_parser, logging_parser, json_output_parser])
    get_netlocalgroup_parser.add_argument('--groupname', dest='queried_groupname',
            help='Group to list the members of (defaults to the local \'Administrators\' group')
    get_netlocalgroup_parser.add_argument('--list-groups', action='store_true',
            help='If set, returns a list of the local groups on the targets')
    get_netlocalgroup_parser.add_argument('-t', '--dc-ip', dest='domain_controller',
            default=str(), help='IP address of the Domain Controller (used to resolve domain SIDs)')
    get_netlocalgroup_parser.add_argument('--tls', action='store_true', dest='do_tls',
            help='Force TLS connection to the Domain Controller')
    get_netlocalgroup_parser.add_argument('-r', '--recurse', action='store_true',
            help='If the group member is a domain group, try to resolve its members as well')
    get_netlocalgroup_parser.set_defaults(func=get_netlocalgroup)

    # Parser for the invoke-checklocaladminaccess command
    invoke_checklocaladminaccess_parser = subparsers.add_parser('invoke-checklocaladminaccess', help='Checks '\
            'if the given user has local admin access on the given host', 
            parents=[target_parser, logging_parser, json_output_parser])
    invoke_checklocaladminaccess_parser.set_defaults(func=invoke_checklocaladminaccess)

    # Parser for the get-netprocess command
    get_netprocess_parser = subparsers.add_parser('get-netprocess', help='This function will '\
        'execute the \'Select * from Win32_Process\' WMI query to a given host for a list of '\
        'executed process', parents=[target_parser, logging_parser, json_output_parser])
    get_netprocess_parser.set_defaults(func=get_netprocess)

    # Parser for the get-userevent command
    get_userevent_parser = subparsers.add_parser('get-userevent', help='This function will '\
        'execute the \'SELECT * from Win32_NTLogEvent\' WMI query to a given host for a list of '\
        'executed process', parents=[target_parser, logging_parser, json_output_parser])
    get_userevent_parser.add_argument('--event-type', nargs='+', choices=['logon', 'tgt'],
            default=['logon', 'tgt'], help='The type of event to search for: logon, tgt, or all (default: all)')
    get_userevent_parser.add_argument('--date-start', type=int,
            default=5, help='(Filter out events before this date (in days) default: %(default)s)')
    get_userevent_parser.set_defaults(func=get_userevent)

    # Parser for the invoke-userhunter command
    invoke_userhunter_parser = subparsers.add_parser('invoke-userhunter', help='Finds '\
            'which machines domain users are logged into', parents=[ad_parser, hunter_parser, logging_parser, json_output_parser])
    invoke_userhunter_parser.add_argument('--unconstrained', action='store_true',
            help='Query only computers with unconstrained delegation')
    invoke_userhunter_parser.add_argument('--admin-count', action='store_true',
            help='Query only users with adminCount=1')
    invoke_userhunter_parser.add_argument('--allow-delegation', action='store_true',
            help='Return user accounts that are not marked as \'sensitive and '\
                    'not allowed for delegation\'')
    invoke_userhunter_parser.add_argument('--check-access', action='store_true',
            help='Check if the current user has local admin access to the target servers')
    invoke_userhunter_parser.add_argument('--stealth', action='store_true',
            help='Only enumerate sessions from commonly used target servers')
    invoke_userhunter_parser.add_argument('--stealth-source', nargs='+', choices=['dfs', 'dc', 'file'],
            default=['dfs', 'dc', 'file'], help='The source of target servers to use, '\
                    '\'dfs\' (distributed file server), \'dc\' (domain controller), '\
                    'or \'file\' (file server) (default: all)')
    invoke_userhunter_parser.add_argument('--foreign-users', action='store_true',
            help='Only return users that are not part of the searched domain')
    invoke_userhunter_parser.add_argument('--stop-on-success', action='store_true',
            help='Stop hunting after finding target')
    invoke_userhunter_parser.add_argument('--show-all', action='store_true',
            help='Return all results')
    invoke_userhunter_parser.set_defaults(func=invoke_userhunter)

    # Parser for the invoke-processhunter command
    invoke_processhunter_parser = subparsers.add_parser('invoke-processhunter', help='Searches machines '\
            'for processes with specific name, or ran by specific users',
            parents=[ad_parser, hunter_parser, json_output_parser, logging_parser])
    invoke_processhunter_parser.add_argument('--processname', dest='queried_processname',
            nargs='+', default=list(), help='Names of the process to hunt')
    invoke_processhunter_parser.add_argument('--stop-on-success', action='store_true',
            help='Stop hunting after finding target')
    invoke_processhunter_parser.add_argument('--show-all', action='store_true',
            help='Return all results')
    invoke_processhunter_parser.set_defaults(func=invoke_processhunter)

    # Parser for the invoke-eventhunter command
    invoke_eventhunter_parser = subparsers.add_parser('invoke-eventhunter', help='Searches machines '\
            'for events with specific name, or ran by specific users',
            parents=[ad_parser, hunter_parser, json_output_parser, logging_parser])
    invoke_eventhunter_parser.add_argument('--search-days', dest='search_days',
            type=int, default=3, help='Number of days back to search logs for (default: %(default)s)')
    invoke_eventhunter_parser.set_defaults(func=invoke_eventhunter)

    args = parser.parse_args()

    # setup the main logger
    logger = logging.getLogger('pywerview_main_logger')
    logging.addLevelName(5, 'ULTRA')
    logger.setLevel(args.logging_level)
    console_handler = logging.StreamHandler()
    console_handler.setLevel(args.logging_level)
    formatter = logging.Formatter('[%(levelname)s] %(name)s - %(funcName)s : %(message)s')
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    if args.hashes:
        try:
            args.lmhash, args.nthash = args.hashes.split(':')
        except ValueError:
            args.lmhash, args.nthash = 'aad3b435b51404eeaad3b435b51404ee', args.hashes
        finally:
            args.password = str()
            if args.lmhash == str():
                args.lmhash = 'aad3b435b51404eeaad3b435b51404ee'
    else:
        args.lmhash = args.nthash = str()

    if (args.password is None and args.hashes is None and not args.do_kerberos 
            and not args.user_cert is None and args.user_key is None):
        from getpass import getpass
        args.password = getpass('Password:')

    parsed_args = dict()
    for k, v in vars(args).items():
        if k not in ('func', 'hashes', 'submodule', 'logging_level', 'json_output'):
            parsed_args[k] = v

    starting_time = datetime.datetime.now()
    results = args.func(**parsed_args)
    ending_time = datetime.datetime.now()

    try:
        json_output = args.json_output
    except AttributeError:
        json_output = False

    if results is not None:
        if json_output:
            results_json = {'cmd' : {'submodule' : args.submodule, 'args' : parsed_args,
                'starting_time': starting_time, 'ending_time': ending_time}}
            try:
                objects_json = [x.to_json() for x in results]
            except TypeError:
                try:
                    objects_json = [results.to_json()]
                except AttributeError:
                    objects_json = results
            results_json['results'] = objects_json
            print(json.dumps(results_json, default=str))
        else:
            try:
                print('\n\n'.join(str(x) for x in results))
            except TypeError:
                print(results)

